"""Helpers for checking if a new CLI version is available."""

import json
from dataclasses import dataclass
from datetime import datetime, timedelta
from pathlib import Path

import requests
from packaging.version import Version
from pytz import UTC
from rich.panel import Panel

from slices_cli_ai import err_console
from slices_cli_ai.const import CONFIG_DIR


@dataclass
class AvailableVersionInfo:
    """Info about the available CLI version in the package repository."""

    stable_version: Version
    pre_version: Version | None
    info_updated_at: datetime


@dataclass
class InstalledVersionInfo:
    """Info about the locally installed CLI version."""

    version: Version

    @property
    def is_prerelease(self) -> bool:
        """Return True if a pre-release installed locally."""
        return self.version.is_prerelease


SLICES_CLI_PACKAGE_INFO_URL = "https://doc.slices-ri.eu/pypi/slices-cli/json"


def download_available_versions() -> AvailableVersionInfo:
    """Contact the package repository, to get info about the available CLI version."""
    response = requests.get(SLICES_CLI_PACKAGE_INFO_URL, timeout=1)
    response.raise_for_status()
    version_info = response.json()
    newest_stable_version = None
    newest_pre_version = None
    for version_str in version_info["releases"]:
        version = Version(version_str)
        if version.is_prerelease:
            if newest_pre_version is None or version > newest_pre_version:
                newest_pre_version = version
        elif newest_stable_version is None or version > newest_stable_version:
            newest_stable_version = version
    assert newest_stable_version is not None
    return AvailableVersionInfo(
        stable_version=newest_stable_version, pre_version=newest_pre_version, info_updated_at=datetime.now(UTC)
    )


def get_running_version() -> InstalledVersionInfo:
    """Return the running Slices CLI ai version info."""
    from importlib.metadata import version as metadata_version  # noqa: PLC0415

    return InstalledVersionInfo(Version(metadata_version("slices-cli")))


config_dir_location = Path.home() / CONFIG_DIR
version_cache_file_location = config_dir_location / "versions_cache.json"
version_warning_date_file_location = config_dir_location / "versions_warning.json"


def save_version_cache(version_info: AvailableVersionInfo):
    """Save the AvailableVersionInfo to the cache."""
    try:
        with version_cache_file_location.open("w") as version_cache_file:
            json.dump(
                {
                    "stable_version": str(version_info.stable_version),
                    "pre_version": str(version_info.pre_version),
                    "info_updated_at": version_info.info_updated_at.timestamp(),
                },
                version_cache_file,
            )
    except:  # noqa: E722, S110
        # ignore
        # raise
        pass


def load_version_cache() -> AvailableVersionInfo | None:
    """Load the AvailableVersionInfo from the cache."""
    if not version_cache_file_location.exists():
        return None
    try:
        with version_cache_file_location.open() as version_cache_file:
            version_cache = json.load(version_cache_file)
            return AvailableVersionInfo(
                stable_version=Version(version_cache["stable_version"]),
                pre_version=Version(version_cache["pre_version"]),
                info_updated_at=datetime.fromtimestamp(version_cache["info_updated_at"], tz=UTC),
            )
    except:  # noqa: E722
        # ignore
        # raise
        return None


def save_version_warning_date(warned_at: datetime):
    """Save the version warning date."""
    try:
        with version_warning_date_file_location.open("w") as version_warning_date_file:
            json.dump({"warned_at": warned_at.timestamp()}, version_warning_date_file)
    except:  # noqa: E722, S110
        # ignore
        pass


def load_version_warning_date() -> datetime | None:
    """Load the version warning date."""
    if not version_warning_date_file_location.exists():
        return None
    try:
        with version_warning_date_file_location.open() as version_warning_date_file:
            version_warning_date = json.load(version_warning_date_file)
            return datetime.fromtimestamp(version_warning_date["warned_at"], tz=UTC)
    except:  # noqa: E722
        # ignore
        return None


UPGRADE_CHECK_INTERVAL = timedelta(hours=1)
UPGRADE_WARN_INTERVAL = timedelta(hours=20)


def is_newer_version_available() -> tuple[InstalledVersionInfo, AvailableVersionInfo] | None:
    """Return version info, or None if no newer version available."""

    available = load_version_cache()
    if available is None or available.info_updated_at < datetime.now(UTC) - UPGRADE_CHECK_INTERVAL:
        available = download_available_versions()
        if available is None:
            return None
        save_version_cache(available)
    installed = get_running_version()
    if installed.is_prerelease:
        if available.pre_version and installed.version < available.pre_version:
            return installed, available
    elif installed.version < available.stable_version:
        return installed, available
    return None


def handle_newer_version_available():
    """Check if new version is available, and output info to console if it is."""
    try:
        has_newer = is_newer_version_available()
        if not has_newer:
            return
        warned_at = load_version_warning_date()
        if warned_at is not None and warned_at > datetime.now(UTC) - UPGRADE_WARN_INTERVAL:
            # Already warned recently
            return
        installed, available = has_newer
        new_v = available.pre_version if installed.is_prerelease else available.stable_version
        if installed.is_prerelease:
            command = "pip install --upgrade --pre slices-cli --extra-index-url=https://doc.slices-ri.eu/pypi/"
        else:
            command = "pip install --upgrade slices-cli --extra-index-url=https://doc.slices-ri.eu/pypi/"
        err_console.print(
            Panel(
                f"""Current version: {installed.version}
    New version: [bold]{new_v}[/bold]
   Upgrade with: [italic]{command}[/italic]""",
                title="[yellow]New Slices CLI ai version available",
                expand=False,
                style="yellow",
                padding=(1, 3),
            )
        )
        save_version_warning_date(datetime.now(UTC))
    except Exception as e:
        # If anything at all goes wrong, ignore and return no new version is available
        err_console.print(f"Problem checking for new version: {e}")
        return



"""Basic infrastructure CLI."""

import functools
import socket

import typer
from slices_clientlib_core.base.exceptions import ApiResponseError
from slices_clientlib_core.experiment import Experiment
from urllib3.connection import HTTPSConnection
from urllib3.exceptions import MaxRetryError, NewConnectionError

from . import err_console
from .ai_context import AIContext
from .utils import is_type_id


def handle_some_exceptions(func):
    """Handle specific exceptions by giving the user a useful error message instead of a confusing stacktrace."""

    @functools.wraps(func)
    def wrapper_handle_some_exceptions(*args, **kwargs):
        """Function wrapper for handle_some_exceptions."""  # noqa: D401
        try:
            return func(*args, **kwargs)
        except ApiResponseError as e:
            if e._httpx_status_error.response.status_code in (
                502,
                503,
                504,
            ):
                err_console.print("[red]There was a connection problem. The service might be down. :zap:")
                err_console.print(f"[red]Details: {e}")
                raise typer.Exit(code=1) from None
            # If it's something else, re-raise the exception and handle as usually
            raise
        except MaxRetryError as e:
            if isinstance(e.reason, NewConnectionError | HTTPSConnection):
                # example:
                #   NewConnectionError -> [Failed to establish a new connection: Errno 113] No route to host
                #   HTTPSConnection -> Failed to establish a new connection: [Errno 111] Connection refused
                err_console.print("[red]There was a connection problem. The service might be down. :zap:")
                err_console.print(f"[red]Details: {e.reason}")
                raise typer.Exit(code=1) from None
            # If it's something else, re-raise the exception and handle as usually
            raise

    return wrapper_handle_some_exceptions


def bi_get_experiment(ctx: AIContext, experiment_id: str) -> Experiment:
    """Get an experiment for a bi command."""
    if is_type_id(experiment_id, type="exp"):
        exp_obj = ctx.obj.api.experiment.get(experiment_id)
    else:
        exp_obj = ctx.obj.api.get_experiment(experiment_id, project_id=ctx.obj.require_project_id())

    if not exp_obj:
        err_console.print(f"Could not retrieve experiment {experiment_id} :x:")
        raise typer.Exit(code=1)
    return exp_obj


def test_ssh_connect(host: str, port: int) -> bool:
    """Check if an SSH connection can be established to the target."""

    # TODO the timeout value is too short. Ideally, it should be based on known RTT of a slices service.

    try:
        with socket.socket(socket.AF_INET6, socket.SOCK_STREAM) as s:
            s.settimeout(0.1)
            if s.connect_ex((host, port)) == 0:
                # if we can connect, return True
                return True
    except:  # noqa: E722 S110
        # ignore IPv6 problems
        pass

    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.settimeout(0.1)
            return s.connect_ex((host, port)) == 0
    except:  # noqa: E722
        # Any exception means we can't connect.
        # Examples:
        #      gaierror: [Errno -5] No address associated with hostname
        return False



"""AI CLI."""

from dataclasses import dataclass
from functools import cached_property

import typer

from . import err_console
from .context import ContextObj


@dataclass
class AIContextObj(ContextObj):
    """AI Context object."""

    site_id: str

    def _require_project_id(self):
        project_id = self.get_current_project_id()
        if not project_id:
            err_console.print("[bold red]No project set. Please set a project with `slices project use <name>`. :x:")
            raise typer.Exit(1)
        return project_id

    @cached_property
    def ai_api(self):
        """AI API."""
        return self.api.for_project(self._require_project_id()).ai(self.site_id)


class AIContext(typer.Context):
    """Custom context for slices-cli."""

    obj: AIContextObj


class WrapContextAsAIContext:
    """Helper to wrap a regular Context so it behaves as a AIContext."""

    def __init__(self, orig_context: typer.Context, ctx_obj: AIContextObj):
        self.orig_context = orig_context
        self.ctx_obj = ctx_obj

    @property
    def obj(self):
        """Return the modified context object."""
        return self.ctx_obj

    def __getattr__(self, attr):
        """Proxy all calls to the real Context."""
        return getattr(self.orig_context, attr)



"""Slices AI CLI."""

import logging
import os

from rich.console import Console

from slices_cli_ai.util.console import DummyConsole, LazyRichConsole

VERBOSITY_TRACE = 3
VERBOSITY_DEBUG = 2
VERBOSITY_INFO = 1


def make_console(*, log_path: bool = True, stderr: bool = False) -> Console:
    """Make a rich console.

    Includes workaround for special case of python via shell process inside a notebook.
    """

    res = Console(log_path=log_path, stderr=stderr)
    if not res.is_jupyter and os.getenv("JPY_SESSION_NAME") is not None:
        # We are running in a `!` shell process inside a notebook.
        # This is a special case.
        #    - This "terminal" looks interactive, but is not.
        #    - This "terminal" understand color codes, but not terminal control codes to move the cursor etc
        # We force rich to see the terminal as DUMB and non-interactive
        # Also, we make sure we don't revert to default dumb terminal size,
        #    which is very small, so we set fallbacks for LINES and COLUMNS
        os.environ["TERM"] = "DUMB"
        return Console(
            log_path=log_path,
            stderr=stderr,
            force_interactive=False,
            width=int(os.getenv("COLUMNS") or 140),
            height=int(os.getenv("LINES") or 30),
        )
    return res


console = make_console(log_path=False)
err_console = make_console(stderr=True)


DEBUG_LEVEL_INFO = 1
DEBUG_LEVEL_DEBUG = 2
DEBUG_LEVEL_TRACE = 3

try:
    DEBUG_LEVEL = int(os.environ.get("SLICES_DEBUG", "0"))
except ValueError:
    DEBUG_LEVEL = 0
    err_console.print("[red bold] :x: ignoring SLICES_DEBUG environment variable: not an integer", highlight=False)

console_info = LazyRichConsole(log_path=False) if DEBUG_LEVEL >= DEBUG_LEVEL_INFO else DummyConsole()
console_debug = LazyRichConsole(log_path=False) if DEBUG_LEVEL >= DEBUG_LEVEL_DEBUG else DummyConsole()
console_trace = LazyRichConsole(log_path=False) if DEBUG_LEVEL >= DEBUG_LEVEL_TRACE else DummyConsole()


class RichConsoleHandler(logging.Handler):
    """Rich console handler."""

    def emit(self, record: logging.LogRecord) -> None:
        """Emit a record."""
        if record.levelno >= logging.WARNING and DEBUG_LEVEL >= DEBUG_LEVEL_INFO:
            console_info.print(f"[yellow bold]{record.getMessage()}", highlight=False)
        elif record.levelno >= logging.INFO and DEBUG_LEVEL >= DEBUG_LEVEL_DEBUG:
            console_info.print(f"[blue]{record.getMessage()}", highlight=False)
        elif record.levelno >= logging.DEBUG and DEBUG_LEVEL >= DEBUG_LEVEL_TRACE:
            console_debug.print(f"[bright_black]{record.getMessage()}", highlight=False)


def handle_logging(logger_name: str):
    """Set up logging for the given logger.

    The logging will be done by using the appropriate console.
    This console can be a rich console or a dummy console depending on the SLICES_DEBUG environment variable.
    """
    slices_clientlib_core_logger = logging.getLogger(logger_name)
    if DEBUG_LEVEL <= DEBUG_LEVEL_TRACE:
        slices_clientlib_core_logger.setLevel(logging.DEBUG)
    elif DEBUG_LEVEL <= DEBUG_LEVEL_DEBUG:
        slices_clientlib_core_logger.setLevel(logging.INFO)
    elif DEBUG_LEVEL <= DEBUG_LEVEL_INFO:
        slices_clientlib_core_logger.setLevel(logging.WARNING)
    slices_clientlib_core_logger.addHandler(RichConsoleHandler())


handle_logging("slices_clientlib_core")
handle_logging("slices_clientlib_aiis")
handle_logging("slices_cli_ai")

# prevent any default logging from seeing our logging
logging.getLogger("slices_cli_ai").propagate = False  # Do not propagate to parent loggers



"""APIs."""

from datetime import datetime

import humanize
import requests
import typer
from pytz import UTC
from slices_clientlib_aiis import SlicesClientWithAIInfraService, SlicesProjectClientWithAIInfraService
from slices_clientlib_aiis.base.config import SlicesAIApiClientsConfig
from slices_clientlib_core import SlicesExperimentClient
from slices_clientlib_core.base.config import SlicesCentralApiClientsConfig
from slices_clientlib_core.base.exceptions import ForbiddenException
from slices_clientlib_core.experiment import Experiment
from slices_clientlib_core.oauth2.config import load_oauth_config
from slices_clientlib_core.oauth2.token import OAuthToken
from slices_clientlib_core.project import Project

from . import console, err_console
from .utils import friendly_datetime


class NotLoggedInException(Exception):
    """Not logged in."""


class ApiClients:
    """API Clients."""

    auth_token: OAuthToken | None

    _slices_client: SlicesClientWithAIInfraService

    def __init__(
        self,
        auth_token: OAuthToken | None,
        central_config: SlicesCentralApiClientsConfig | str | None = None,
        ai_config: SlicesAIApiClientsConfig | str | None = None,
    ):
        """Create API clients."""
        self.auth_token = auth_token
        self.oauth_config = load_oauth_config()

        self._slices_client = SlicesClientWithAIInfraService(central_config, ai_config)

    @property
    def central_config(self) -> SlicesCentralApiClientsConfig:
        """Return the Slices central config object."""
        return self._slices_client._central_api_clients_config

    @property
    def ai_config(self) -> SlicesAIApiClientsConfig:
        """Return the Slices Basic Infrastructure config object."""
        return self._slices_client.ai_api_clients_config

    # TODO can we prevent the copy paste here?

    def reload_ai_config(self) -> bool:
        """Reload the Slices AI Infrastructure Service config object from the default online source.

        This is useful because it is locally cached.

        This does nothing if the bi slices config was not loaded from
        the default online source or its cached version.

        Returns: True if reload was done, False if reload was skipped.
        """
        return self._slices_client.reload_ai_config()

    @property
    def public_key(self):
        """Pubkey API."""
        return self._slices_client.public_key

    @property
    def experiment(self):
        """Experiment API."""
        return self._slices_client.experiment

    @property
    def project(self):
        """Project API."""
        return self._slices_client.project

    @property
    def user(self):
        """User API."""
        return self._slices_client.user

    def for_project(self, project_id: str) -> SlicesProjectClientWithAIInfraService:
        """Slices Project Client."""
        return self._slices_client.for_project(project_id)

    def for_experiment(self, experiment_id: str, project_id: str) -> SlicesExperimentClient:
        """Slices Experiment Client."""
        return self._slices_client.for_experiment(experiment_id, project_id)

    def get_id_token_for_audience(self, audience: str | list[str], project_id: str):
        """Return JWT ID token of logged in user."""
        if not self.auth_token:
            raise NotLoggedInException

        data = {
            "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
            "subject_token": self.auth_token.access_token,
            "subject_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "client_id": self.oauth_config.client_id,
            "client_secret": self.oauth_config.client_secret,
            "requested_token_type": "urn:ietf:params:oauth:token-type:id_token",
            "audience": audience,
            "scope": f"{self.auth_token.scope} project:{project_id}",
        }

        response = requests.post(
            url=self.oauth_config.token_endpoint_url,
            data=data,
            timeout=10,
        )
        response.raise_for_status()
        return response.json()["access_token"]

    def get_access_token_for_audience(self, audience: str | list[str], project_id: str | None = None):
        """Return JWT access token of logged in user for target resource."""
        if not self.auth_token:
            raise NotLoggedInException

        scope = self.auth_token.scope
        if project_id:
            scope = f"{scope} project:{project_id}"

        data = {
            "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
            "subject_token": self.auth_token.access_token,
            "subject_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "client_id": self.oauth_config.client_id,
            "client_secret": self.oauth_config.client_secret,
            "requested_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "audience": audience,
            "scope": scope,
        }

        response = requests.post(
            url=self.oauth_config.token_endpoint_url,
            data=data,
            timeout=10,
        )
        response.raise_for_status()
        return response.json()["access_token"]

    def get_experiment(self, experiment_name: str, *, project_id: str, must_be_active=False) -> Experiment | None:
        """Get experiment by name."""
        return next(
            iter(
                self.experiment.list(
                    friendly_name=experiment_name,
                    project_id=project_id,
                    include_archived=not must_be_active,
                    include_expired=not must_be_active,
                )
            ),
            None,
        )

    def get_experiment_by_id(self, experiment_id: str) -> Experiment:
        """Get Experiment by ID."""
        return self.experiment.get(experiment_id)

    def get_or_create_experiment(self, experiment_name: str, project_id: str, expires_at: datetime) -> Experiment:
        """Get or create experiment by name."""
        experiment = self.get_experiment(
            experiment_name,
            project_id=project_id,
            must_be_active=True,
        )

        if not experiment:
            # Create it

            user_project_id_token = self.get_id_with_project_jwt(
                self.central_config.central_hub_services.experiment_authority.audience, project_id
            )

            # First make sure project has not yet expired.
            project = self.project.get(project_id)
            if project and project.expires_at and project.expires_at < datetime.now(UTC):
                hum_ago = humanize.precisedelta(
                    project.expires_at - datetime.now(UTC), minimum_unit="hours", format="%d"
                )
                console.print(
                    f"[red]:clock10: Project expired on {friendly_datetime(project.expires_at)} ({hum_ago} ago)."
                )
                raise typer.Exit(code=1) from None

            experiment = self.experiment.create(
                friendly_name=experiment_name,
                description=None,
                expires_at=expires_at,
                user_project_id_token=user_project_id_token,
            )

        return experiment

    def get_experiment_jwt(self, experiment_id: str) -> str:
        """Get Experiment JWT."""
        try:
            return self.experiment.get_jwt(experiment_id)
        except ForbiddenException as err:
            err_console.print(
                f"[red]You have insufficient permissions to create a token for experiment [i]{experiment_id}[/] . :x:"
            )
            raise typer.Exit(1) from err

    def get_project_jwt(self, project_id: str) -> str:
        """Get Project JWT."""

        try:
            return self.project.get_jwt(project_id)
        except ForbiddenException as err:
            err_console.print(
                f"[red]You have insufficient permissions to create a token for project [i]{project_id}[/] . :x:"
            )
            raise typer.Exit(1) from err

    def get_id_with_project_jwt(self, audience: str, project_id: str) -> str:
        """Get ID Token with project JWT."""

        return self.get_id_token_for_audience(audience, project_id)

    def get_project_by_name(self, project_name: str) -> Project | None:
        """Get project by name."""
        for project in self.project.list():
            if project_name in (project.id, project.name):
                return project
        return None



"""Custom slices-cli context."""

import os
from dataclasses import dataclass
from typing import Literal

import typer
from slices_clientlib_aiis.base.config import SlicesAIApiClientsConfig
from slices_clientlib_core.base.config import SlicesCentralApiClientsConfig, get_config_string
from slices_clientlib_core.oauth2.token import AuthJWT
from typer import Context as BaseContext

from . import err_console
from .api import ApiClients
from .utils import is_project_id


@dataclass
class ContextObj:
    """Context Object."""

    api: ApiClients
    user_info: AuthJWT | None
    verbosity: int
    project_id: str | None

    @property
    def central_config(self) -> SlicesCentralApiClientsConfig:
        """Return the Slices central config object."""
        return self.api.central_config

    @property
    def ai_config(self) -> SlicesAIApiClientsConfig:
        """Return the Slices BI config object."""
        return self.api.ai_config

    def get_logged_in_user(self) -> AuthJWT:
        """Get user info or show error."""
        if not self.user_info:
            err_console.print("[red]You must be logged in to perform this action. :key:")
            raise typer.Exit(128)
        return self.user_info

    def get_logged_in_userinfo(self, field_name: Literal["sub", "preferred_username"]):
        """Return logged in user info."""
        return self.get_logged_in_user()[field_name]

    def get_current_project_id(self) -> str | None:
        """Return the current project."""

        if project_id := os.environ.get("SLICES_PROJECT_ID"):
            if not is_project_id(project_id):
                err_console.print(
                    f"[red]Invalid project id '{project_id}' passed via environment variable 'SLICES_PROJECT_ID' :x:"
                )
            return os.environ["SLICES_PROJECT_ID"]

        project_id = get_config_string("default", section="project")

        if project_id and not is_project_id(project_id):
            err_console.print(f"[red]Invalid project id '{project_id}' :x:")
            err_console.print(
                "[red]Please check your configuration. "
                "You can set a valid project with [blue]slices project use <project name>[/]"
            )
            raise typer.Exit(130)
        return project_id

    def require_project_id(self) -> str:
        """Require a current project.

        Print an error and exit if not available.
        """
        project_id = self.get_current_project_id()
        if not project_id:
            err_console.print("[red]You must select a project to perform this action. :key:")
            err_console.print("Hint: Set a project with [blue]slices project use <project name>[/]")
            raise typer.Exit(129)

        return project_id

    def require_project_friendly_name(self) -> str:
        """Require a current project.

        Print an error and exit if not available.

        Return the friendly name of the project.
        """
        project_id = self.get_current_project_id()
        if not project_id:
            err_console.print("[red]You must select a project to perform this action. :key:")
            err_console.print("Hint: Set a project with [blue]slices project use <project name>[/]")
            raise typer.Exit(129)

        return self.api.project.get(project_id).name

    def get_current_project_token(self) -> str | None:
        """Return token for current project."""
        project_id = self.get_current_project_id()
        if not project_id:
            return None

        return self.api.get_project_jwt(project_id)


class Context(BaseContext):
    """Custom context for slices-cli."""

    obj: ContextObj



"""AI Infrastructure Service CLI - Query jobs."""

import re
from datetime import datetime, timedelta
from enum import Enum
from typing import TYPE_CHECKING, Annotated, Any

import typer
from msgspec.json import encode as json_encode
from msgspec.yaml import encode as yaml_encode
from pytz import UTC
from rich.table import Column, Table
from slices_clientlib_aiis.schema.job import (
    Job,
    JobPortMapping,
    JobStatus,  # noqa: TCH002 required outside type checking! ruff is wrong.
)
from slices_clientlib_aiis.schema.job_debug_logs import JobDebugLogType

from . import console
from .ai import ai_app
from .ai_common import get_ai_job, get_requested_ai_jobs, project_id_to_name, status_to_color, user_id_to_name
from .ai_context import AIContext
from .ai_helpers import (
    handle_some_exceptions,
)
from .utils import friendly_datetime, td_format

if TYPE_CHECKING:
    from collections.abc import Collection


class ListResourcesFormat(Enum):
    """List Resources Formats."""

    table = "table"
    json = "json"
    yaml = "yaml"


class JobOwnerSelection(Enum):
    """Show jobs with which owner."""

    user = "user"
    project = "project"
    all = "all"


def _get_cluster_id(job: Job) -> int | None:
    """Get cluster ID from job, or return None if not available."""
    if job.state.resources is not None:
        return job.state.resources.cluster_id
    if job.state.scheduling is not None and job.state.scheduling.assigned_cluster_id is not None:
        return job.state.scheduling.assigned_cluster_id
    if job.request.resources.cluster_id is not None:
        return job.request.resources.cluster_id
    return None


@ai_app.command("list", rich_help_panel="Information commands")
@handle_some_exceptions
def list_jobs(
    ctx: AIContext,
    job_ids: Annotated[list[str] | None, typer.Argument(help="Zero, one or more job IDs.", show_default=False)] = None,
    include_pending: Annotated[
        bool, typer.Option("--show-pending/--hide-pending", help="Show/hide QUEUED and ONHOLD jobs.")
    ] = True,
    include_running: Annotated[
        bool, typer.Option("--show-running/--hide-running", help="Show/hide jobs in any running state.")
    ] = True,
    include_finished: Annotated[
        bool, typer.Option("--show-finished/--hide-finished", help="Show/hide jobs that have stopped.")
    ] = True,
    include_deleted: Annotated[
        bool, typer.Option("--show-deleted/--hide-deleted", help="Show/hide deleted jobs.")
    ] = False,
    format: Annotated[
        ListResourcesFormat,
        typer.Option("--format", "-f", help="Format in which the resources should be listed."),
    ] = "table",  # type: ignore
    filter_owner: Annotated[
        JobOwnerSelection,
        typer.Option("--owner-filter", "-o", help="Show only your own jobs, those in your project, or all jobs."),
    ] = JobOwnerSelection.project,
    limit: Annotated[
        int,
        typer.Option("--limit", help="Show only a limited number of jobs.", min=1, max=500),
    ] = 20,
):
    """List your jobs."""

    jobs_to_show: Collection[Job] = get_requested_ai_jobs(
        ctx,
        job_ids,
        filter_user=filter_owner == JobOwnerSelection.user,
        filter_project=filter_owner == JobOwnerSelection.project,
        include_pending=include_pending,
        include_running=include_running,
        include_finished=include_finished,
        include_deleted=include_deleted,
    )

    total_job_count = len(jobs_to_show)
    if total_job_count > limit:
        limited_jobs_to_show = []
        for j in jobs_to_show:
            limited_jobs_to_show.append(j)
            if len(limited_jobs_to_show) == limit:
                break
        jobs_to_show = limited_jobs_to_show

    if format == ListResourcesFormat.table:
        cols = [
            "ID",
            "Name",
            "Status",
            "Created At",
            "Cluster ID",
            "User",
            "Project",
        ]
        title = "Jobs"
        if filter_owner == JobOwnerSelection.user:
            title = "Your Jobs"
        elif filter_owner == JobOwnerSelection.project:
            title = f"Jobs in Project {ctx.obj.require_project_friendly_name()}"
        title += f" ({len(jobs_to_show)}/{total_job_count})"
        table = Table(*cols, title=title)
        for j in jobs_to_show:
            cluster_id = _get_cluster_id(j)
            rows = [
                str(j.id),
                j.name,
                f"{status_to_color(j.state.status)}{j.state.status.name}",
                friendly_datetime(j.state.event_times.created.astimezone()),
                str(cluster_id) if cluster_id else "-",
                user_id_to_name(ctx, j.owner.user_id),
                project_id_to_name(ctx, j.owner.project_id),
            ]
            table.add_row(*rows)
        console.print(table)

    elif format == ListResourcesFormat.json:
        console.print_json(json_encode(list(jobs_to_show)).decode())
    elif format == ListResourcesFormat.yaml:
        console.print(yaml_encode(list(jobs_to_show)).decode())


class ShowResourceFormat(Enum):
    """Show Resource Formats."""

    table = "table"
    json = "json"
    yaml = "yaml"


DOCKER_IMAGE_PATTERN = (
    r"^((?P<username>[\w\-+_.]+):(?P<password>[^@ ]+)@)?"  # passwords with space or @ are not supported
    r"(?P<repository>[\w.\-_]+(?::\d+|)\/)?"
    r"(?P<image>[a-z0-9.\-_]+(?:\/[a-z0-9.\-_]+|)+)(:(?P<tag>[\w.\-_]{1,127}))?$"
)

DOCKER_IMAGE_REGEX = re.compile(DOCKER_IMAGE_PATTERN)


@ai_app.command("show", rich_help_panel="Job interaction commands")
@handle_some_exceptions
def show_job(  # noqa: PLR0915, C901
    ctx: AIContext,
    job_id: Annotated[str, typer.Argument(help="job ID", show_default=False)],
    format: Annotated[
        ShowResourceFormat,
        typer.Option("--format", "-f", help="Format in which the job should be shown."),
    ] = "table",  # type: ignore
):
    """Show Job Details."""

    job: Job = get_ai_job(ctx, job_id)

    if format == ShowResourceFormat.table:

        def handle_none(i: Any | None) -> Any:
            if i is None:
                return "-"
            return i

        def image_nopass(image: str | None) -> str | None:
            """Docker image with the password removed."""
            if not image:
                return None
            match = DOCKER_IMAGE_REGEX.match(image)

            if match and match.group("password"):
                return (
                    f"{match.group('username')}:XXXX@"
                    f"{match.group('repository')}{match.group('image')}:{match.group('tag')}"
                )

            return image

        def command_as_str(command: str | list | None) -> str | None:
            """Command as a string."""
            if not command:
                return None
            if isinstance(command, str):
                return command
            return " ".join(command)

        def _ssh_port_command(port: int | None) -> str:
            if port and port != 22:  # noqa: PLR2004
                return f" -oPort={port}"
            return ""

        def _port_mapping_format(port_mappings: list[JobPortMapping]) -> str:
            try:
                res = [f"{port_mapping.container_port} -> {port_mapping.host_port}" for port_mapping in port_mappings]
                return ", ".join(res)
            except:  # noqa: E722
                return f"Error processing port mappings: {port_mappings}"

        table = Table.grid(Column(header="Key", justify="right", style="bold"), "Value", padding=(0, 1, 0, 1))

        table.add_row("Job ID:", str(job.uuid))
        table.add_row("Name:", handle_none(job.name))
        table.add_row("Description:", handle_none(job.description))
        table.add_row("Project:", handle_none(job.owner.project_id) if job.owner else None)
        table.add_row("User ID:", handle_none(job.owner.user_id) if job.owner else None)
        table.add_row("Docker image:", handle_none(image_nopass(job.request.docker.image)))
        table.add_row("Command:", handle_none(command_as_str(job.request.docker.command)))

        table.add_row(
            "Status:",
            f"{status_to_color(job.state.status)}{job.state.status.name}" if job.state and job.state.status else "-",
        )

        if job.state == JobStatus.FAILED:
            table.add_row("", f"To print more info about FAILED: slices ai debug {job.uuid}")

        table.add_row(
            "Cluster ID:",
            str(handle_none(job.state.resources.cluster_id)) if job.state and job.state.resources else "-",
        )
        # No longer useful info:
        # table.add_row(
        #     "Worker ID:",
        #     str(handle_none(job.state.resources.worker_id)) if job.state and job.state.resources else "-"
        # )
        table.add_row(
            "Machine ID:", handle_none(job.state.resources.machine_id) if job.state and job.state.resources else "-"
        )
        if job.state.status == JobStatus.QUEUED:
            table.add_row(
                "Queued because:",
                handle_none(" \n".join(job.state.scheduling.queued_explanations))
                if job.state and job.state.scheduling
                else "-",
            )

        if job.state and job.state.resources:
            if job.state.resources.slave_name:
                table.add_row(
                    "Worker Name:",
                    handle_none(job.state.resources.slave_name if job.state and job.state.resources else None),
                )
            if job.state.resources.port_mappings:
                table.add_row(
                    "Port Mappings:",
                    handle_none(
                        _port_mapping_format(job.state.resources.port_mappings)
                        if job.state and job.state.resources
                        else None
                    ),
                )
                if job.state.resources.slave_host:
                    table.add_row(
                        "Worker Host:",
                        handle_none(job.state.resources.slave_host if job.state and job.state.resources else None),
                    )
            if job.state.resources.ssh_username and job.state.resources.ssh_host and job.state.resources.ssh_port:
                ssh_login_str = (
                    f"ssh{_ssh_port_command(job.state.resources.ssh_port)} "
                    f"{job.state.resources.ssh_username}@{job.state.resources.ssh_host}"
                )
                if job.state.resources.ssh_proxy_host:
                    ssh_proxy_login_str = (
                        f' -oProxyCommand="ssh{_ssh_port_command(job.state.resources.ssh_proxy_port)} '
                        f'{job.state.resources.ssh_proxy_username}@{job.state.resources.ssh_proxy_host} -W %h:%p"'
                    )
                    ssh_login_str += ssh_proxy_login_str
                table.add_row("SSH Login:", ssh_login_str)

        timing_table = Table.grid(Column(header="Key", justify="right", style="bold"), "Value", padding=(0, 1, 0, 1))
        if job.state:

            def _show_rel(shown_dt: datetime | None, earlier_dt: datetime | None, after_subject) -> str:
                if not shown_dt or not earlier_dt:
                    return ""
                rel_str = td_format(shown_dt - earlier_dt)
                return f" ({rel_str} after {after_subject})"

            def _show_ago(shown_dt: datetime | None) -> str:
                if not shown_dt:
                    return ""
                rel_str = td_format(datetime.now(UTC) - shown_dt)
                return f" ({rel_str} ago)"

            def _date_to_str(dt: datetime | None) -> str | None:
                if not dt:
                    return None
                used_dt = dt - timedelta(microseconds=dt.microsecond)
                # change timezone to local
                used_dt = used_dt.astimezone(tz=None)
                return used_dt.isoformat()

            if job.state.event_times:
                timing_table.add_row("", "[bold]Timing:")
                timing_table.add_row(
                    "Created:",
                    handle_none(_date_to_str(job.state.event_times.created)) + _show_ago(job.state.event_times.created),
                )
                timing_table.add_row(
                    "Queued:",
                    handle_none(_date_to_str(job.state.event_times.QUEUED))
                    + _show_rel(job.state.event_times.QUEUED, job.state.event_times.created, "job creation"),
                )
                timing_table.add_row(
                    "Assigned:",
                    handle_none(_date_to_str(job.state.event_times.ASSIGNED))
                    + _show_rel(job.state.event_times.ASSIGNED, job.state.event_times.QUEUED, "QUEUED"),
                )
                timing_table.add_row(
                    "Starting:",
                    handle_none(_date_to_str(job.state.event_times.STARTING))
                    + _show_rel(job.state.event_times.STARTING, job.state.event_times.ASSIGNED, "ASSIGNED"),
                )
                timing_table.add_row(
                    "Running:",
                    handle_none(_date_to_str(job.state.event_times.RUNNING))
                    + _show_rel(job.state.event_times.RUNNING, job.state.event_times.STARTING, "STARTING"),
                )

                if job.state.event_times.MUSTHALT and job.state.event_times.HALTING:
                    timing_table.add_row(
                        "HALT request:",
                        handle_none(_date_to_str(job.state.event_times.MUSTHALT))
                        + _show_rel(job.state.event_times.MUSTHALT, job.state.event_times.RUNNING, "RUNNING"),
                    )
                    timing_table.add_row(
                        "HALT procedure start:",
                        handle_none(_date_to_str(job.state.event_times.HALTING))
                        + _show_rel(job.state.event_times.HALTING, job.state.event_times.MUSTHALT, "HALT request"),
                    )
                    if job.state.event_times.HALTED:
                        timing_table.add_row(
                            "HALT success:",
                            handle_none(_date_to_str(job.state.event_times.HALTED))
                            + _show_rel(
                                job.state.event_times.HALTED, job.state.event_times.HALTING, "HALT procedure start"
                            ),
                        )
                    if job.state.event_times.FAILED:
                        timing_table.add_row(
                            "HALT failure:",
                            handle_none(_date_to_str(job.state.event_times.FAILED))
                            + _show_rel(
                                job.state.event_times.FAILED, job.state.event_times.HALTING, "HALT procedure start"
                            ),
                        )
                timing_table.add_row(
                    "Ended:",
                    handle_none(_date_to_str(job.state.event_times.end_date))
                    + _show_ago(job.state.event_times.end_date),
                )
                duration = job.state.event_times.duration_or_none
                timing_table.add_row("Duration:", td_format(duration) if duration else "-")
                timing_table.add_row(
                    "State Updated:",
                    handle_none(_date_to_str(job.state.event_times.status_updated))
                    + _show_ago(job.state.event_times.status_updated),
                )

            console.print(table)
            if timing_table.columns:
                console.print("")
                # console.print("\n[bold]Timing:")
                console.print(timing_table)
    elif format == ShowResourceFormat.json:
        console.print_json(json_encode(job).decode())
    elif format == ShowResourceFormat.yaml:
        console.print(yaml_encode(job).decode(), highlight=False)


@ai_app.command("output", rich_help_panel="Job interaction commands")
@handle_some_exceptions
def get_job_output(
    ctx: AIContext,
    job_id: Annotated[str, typer.Argument(help="Job ID.")],
):
    """Show Job output."""

    job = get_ai_job(ctx, job_id)
    logs = ctx.obj.api.for_project(ctx.obj.require_project_id()).ai(ctx.obj.site_id).jobs.output.get(job.id)
    # output without markup
    typer.echo(logs)


class DebugFormat(Enum):
    """Job Debug Info Formats."""

    table = "table"
    json = "json"
    yaml = "yaml"


@ai_app.command("debug", rich_help_panel="Auxiliary commands")
@handle_some_exceptions
def get_job_debug(
    ctx: AIContext,
    job_id: Annotated[str, typer.Argument(help="Job ID.")],
):
    """Show internal event log for a job."""

    def _4charsto5(v: str) -> str:
        return " " + v if len(v) == 4 else v  # noqa: PLR2004

    job = get_ai_job(ctx, job_id)
    debug_output = ctx.obj.api.for_project(ctx.obj.require_project_id()).ai(ctx.obj.site_id).jobs.logs.get(job.id)
    for out in debug_output:
        msg = (
            f"[blue]{_4charsto5(out.type.name)}[/blue]: {out.msg}"
            if out.type != JobDebugLogType.STATUS_CHANGE
            else f"[green bold]Status to {out.new_state.name if out.new_state else '?'}"
        )
        console.print(f"[bright_black]{out.time}[/bright_black]: {msg}", highlight=False)


# TODO: add
# get_slave_info
# get_info
# wait



"""AI Infrastructure Service CLI - Cluster info."""

from enum import Enum
from typing import Annotated

import typer
from msgspec import to_builtins
from msgspec.json import encode as json_encode
from msgspec.yaml import encode as yaml_encode
from rich.table import Table
from slices_clientlib_aiis.schema.worker import GpuModel

from . import console
from .ai import ai_app
from .ai_context import AIContext
from .ai_helpers import (
    handle_some_exceptions,
)


class ListClustersFormat(Enum):
    """List Clusters Formats."""

    table = "table"
    json = "json"
    yaml = "yaml"


@ai_app.command("clusters", rich_help_panel="Information commands")
@handle_some_exceptions
def list_clusters(
    ctx: AIContext,
    cluster_id: Annotated[str | None, typer.Argument(help="ID of the cluster to show info for.")] = None,
    all: Annotated[bool, typer.Option("--all", "-a", help="(if no <cluster_id>) Show all machine info.")] = False,
    gpu_details: Annotated[bool, typer.Option("--gpu-details", help="Show more details about GPUs")] = False,
    format: Annotated[
        ListClustersFormat,
        typer.Option("--format", "-f", help="Format in which the resources should be listed."),
    ] = "table",  # type: ignore
):
    """Show available clusters and their resources."""

    cluster_info = ctx.obj.api.for_project(ctx.obj.require_project_id()).ai(ctx.obj.site_id).workers.list_clusters()

    if format in (ListClustersFormat.json, ListClustersFormat.yaml):
        if cluster_id or all:
            # List one or all clusters, with machine info
            machine_info = (
                ctx.obj.api.for_project(ctx.obj.require_project_id()).ai(ctx.obj.site_id).workers.list_slaves()
            )

            clusters_with_machines = []
            for cluster in cluster_info:
                if cluster_id and str(cluster.cluster_id) != str(cluster_id):
                    continue

                cluster_dict = to_builtins(cluster)
                cluster_dict["machines"] = [
                    to_builtins(machine)
                    for machine in machine_info
                    if str(machine.cluster_id) == str(cluster.cluster_id)
                ]
                clusters_with_machines.append(cluster_dict)

            if format == ListClustersFormat.json:
                console.print_json(json_encode(clusters_with_machines).decode())
            elif format == ListClustersFormat.yaml:
                console.print(yaml_encode(clusters_with_machines).decode())
        elif format == ListClustersFormat.json:
            console.print_json(json_encode(list(cluster_info)).decode())
        elif format == ListClustersFormat.yaml:
            console.print(yaml_encode(list(cluster_info)).decode())
        return

    def _gpu_str(gpu: GpuModel) -> str:
        if gpu_details:
            return f"{gpu.name} {gpu.vendor} ({gpu.memory_mb / 1024.0:.2f}GB)"
        return gpu.name

    if cluster_id or all:
        # List one or all clusters, with machine info
        machine_info = ctx.obj.api.for_project(ctx.obj.require_project_id()).ai(ctx.obj.site_id).workers.list_slaves()

        for cluster in cluster_info:
            if cluster_id and str(cluster.cluster_id) != str(cluster_id):
                continue

            console.print(f"[bold]Cluster {cluster.cluster_id}[/bold] - {cluster.comment}")
            cols = [
                "Machine",
                "GPU Model",
                "GPUs",
                "CPUs",
                "Memory (GB)",
                "Accepting Jobs",
                "Active Jobs",
                "CUDA version",
                "CPU model",
            ]
            table = Table(*cols, title=f"Machines in Cluster {cluster.cluster_id}")
            for machine in machine_info:
                if str(machine.cluster_id) == str(cluster.cluster_id):
                    table.add_row(
                        machine.name,
                        ",".join([_gpu_str(g) for g in cluster.gpu_model]),
                        f"{machine.gpu.available}/{machine.gpu.acquired}",
                        f"{machine.cpu.available}/{machine.cpu.acquired}",
                        f"{machine.cpu_memory_mb.available / 1000:.2f}/{machine.cpu_memory_mb.acquired / 1000:.2f}",
                        str(machine.accepting_jobs),
                        str(len(machine.active_job_uuids)) if machine.active_job_uuids is not None else "-",
                        machine.cuda_version_full,
                        ",".join(machine.cpu_model),
                    )
            console.print(table)
    else:
        # List all clusters, without any machine info
        cols = ["ID", "GPU Model", "Comment", "Machines (count)", "GPUs", "CPUs", "Memory (GB)"]

        table = Table(*cols, title="Clusters")
        for cluster in cluster_info:
            table.add_row(
                str(cluster.cluster_id),
                ",".join([_gpu_str(g) for g in cluster.gpu_model]),
                cluster.comment,
                str(cluster.slave_count),
                f"{cluster.gpu.available}/{cluster.gpu.acquired}",
                f"{cluster.cpu.available}/{cluster.cpu.acquired}",
                f"{cluster.cpu_memory_mb.available / 1000:.2f}/{cluster.cpu_memory_mb.acquired / 1000:.2f}",
            )
        console.print(table)



"""AI Infrastructure Service CLI - Create Job."""

import logging
from datetime import datetime, timedelta
from typing import Annotated

import pyjson5
import typer
from msgspec import ValidationError
from msgspec.json import decode as json_decode
from msgspec.json import encode as json_encode
from pyjson5 import Json5DecoderException
from pytz import UTC
from slices_clientlib_aiis.schema.job import JobStatus
from slices_clientlib_aiis.schema.new_job import NewJob
from slices_clientlib_core.base.exceptions import (
    ApiException,
    ForbiddenException,
)
from slices_clientlib_core.experiment import Experiment

from . import console, err_console
from .ai import ai_app
from .ai_common import duration_string_to_seconds
from .ai_context import AIContext
from .ai_helpers import (
    handle_some_exceptions,
)
from .ai_ssh_pubkey import SshKeyParamType, get_ssh_keys
from .ai_subcommands_wait import wait_for_job_status
from .utils import get_api_error_detail

logger = logging.getLogger(__name__)


def _submit_new_job_request(ctx: AIContext, new_job: NewJob, wait_run: bool, wait_done: bool, exp: Experiment):
    """Submit NewResourcesRequest to AI service.

    Optionally wait for the job to be running or finished.

    Show job_id in output.
    """
    try:
        created_job = (
            ctx.obj.api.for_project(ctx.obj.require_project_id())
            .for_experiment(exp.id)
            .ai(ctx.obj.site_id)
            .jobs.create(new_job)
        )
    except ForbiddenException as err:
        err_console.print("[bold red]No permission to create job[/] :x:")
        err_console.print(f"[red]{get_api_error_detail(err)}")
        raise typer.Exit(code=1) from None
    except ApiException as err:
        err_console.print("[bold red]Could not create job[/] :x:")
        err_console.print(f"[red]{get_api_error_detail(err)}")
        raise typer.Exit(code=1) from None

    if wait_run or wait_done:
        wait_for_job_status(ctx, created_job.id, JobStatus.RUNNING if wait_run else JobStatus.FINISHED)

    console.print(f"[bold green]:sparkles: Created Job {created_job.id}")


def _get_or_create_experiment(
    ctx: AIContext, project_id: str, experiment_name: str, expires_at: timedelta
) -> Experiment:
    now = datetime.now(UTC)
    slices_client = ctx.obj.api._slices_client
    user_project_id_token = slices_client.get_id_token(
        slices_client._central_api_clients_config.central_hub_services.experiment_authority.audience, project_id
    )
    if experiment_name.startswith("ai.slices-ri.eu"):
        description = f"ai.slices-ri.eu job '{experiment_name[len('ai.slices-ri.eu-') :]}'"
    else:
        description = None
    return ctx.obj.api.experiment.get_or_create(
        experiment_name, project_id, now + expires_at + timedelta(minutes=2), user_project_id_token, description
    )


def get_job_max_end_time_s(job_max_duration: str | None, job_not_after: datetime | str | None) -> int:
    """Return the maximum end time in seconds from now for the job."""
    now = datetime.now(UTC)
    max_duration_str = job_max_duration or "todo default"
    max_duration_s = duration_string_to_seconds(max_duration_str)
    if max_duration_s is None:
        err_console.print(
            f"[bold red]request.scheduling.max_duration is not a valid duration: {max_duration_str}[/] :x:"
        )
        raise typer.Exit(code=1)

    not_after_val = job_not_after or "1 week"
    if isinstance(not_after_val, str):
        not_after_duration_s = duration_string_to_seconds(not_after_val)

        if not_after_duration_s is None:
            # assume datetime string: parse iso date
            try:
                # note: before python 3.11, fromisoformat did not support 'Z' for UTC
                not_after_date = datetime.fromisoformat(not_after_val.replace("Z", "+00:00"))
            except ValueError:
                err_console.print(
                    f"[bold red]request.scheduling.not_after is not a valid date or duration: {not_after_val}[/] :x:"
                )
                raise typer.Exit(code=1) from None
            not_after_s = int((not_after_date - now).total_seconds())
        else:
            not_after_s = not_after_duration_s
    elif isinstance(not_after_val, datetime):
        # This won't occur because msgspec decodes to str. See NewJobRequestScheduling.
        not_after_s = int((not_after_val - now).total_seconds())
    else:
        err_console.print(f"[bold red]request.scheduling.not_after is not valid: {not_after_val!r}[/] :x:")
        raise typer.Exit(code=1)
    return not_after_s + max_duration_s


@ai_app.command(rich_help_panel="Lifecycle commands")
@handle_some_exceptions
def submit(
    ctx: AIContext,
    file: Annotated[typer.FileText, typer.Argument(..., help="File with job definition.", show_default=False)],
    wait_run: Annotated[
        bool, typer.Option("--wait-run", help="Wait till the job is running.", show_default=False)
    ] = False,
    wait_done: Annotated[
        bool, typer.Option("--wait-done", help="Wait till the job is done.", show_default=False)
    ] = False,
    ssh_key: Annotated[
        list[str] | None,
        typer.Option(
            help="(Extra) SSH public key to register for login to the created job.",
            show_default=False,
            envvar="SLICES_SSH_KEY",
            click_type=SshKeyParamType(),
        ),
    ] = None,
    ssh_key_file: Annotated[
        list[typer.FileText] | None,
        typer.Option(
            help="(Extra) SSH public key file for login to the created job.",
            show_default=False,
            envvar="SLICES_SSH_KEY_FILE",
        ),
    ] = None,
    experiment: Annotated[
        str | None, typer.Option("--experiment", help="Name of the experiment (default: job.name).", show_default=False)
    ] = None,
):
    """Submit a job described in a file."""
    request = _read_and_process_job_request_file(ctx, file)
    ssh_authorized_keys = get_ssh_keys(ctx, ssh_key, ssh_key_file)
    combined_ssh_authorized_keys = set(request.request.extra.ssh_pub_keys)
    combined_ssh_authorized_keys.update(ssh_authorized_keys)  # TODO use smart join of SSH keys (ignore comments)
    request.request.extra.ssh_pub_keys = list(combined_ssh_authorized_keys)

    project_id = ctx.obj.require_project_id()
    expires_at = timedelta(
        seconds=get_job_max_end_time_s(
            request.request.scheduling.max_duration, request.request.scheduling.not_after or "5 min"
        )
    )
    experiment_name = experiment if experiment else f"ai.slices-ri.eu {request.name[:30]}"
    exp = _get_or_create_experiment(ctx, project_id, experiment_name, expires_at)

    _submit_new_job_request(ctx, request, wait_run, wait_done, exp)


def _read_and_process_job_request_file(_ctx: AIContext, file: typer.FileText) -> NewJob:
    """Return NewJob request."""
    try:
        # We support JSON5 instead of plain JSON, which allows trailing comma's, comments, etc.
        # See https://json5.org/
        request_dict = pyjson5.load(file)  # pyright: ignore [reportArgumentType] FileText is compatible enough it seems
    except Json5DecoderException as e:
        err_console.print(f"[red]Errors parsing JSON5 in request file: {e}")
        raise typer.Exit(code=1) from e
    except OSError as e:
        err_console.print(f"[red]Error reading request file: {type(e)}: {e}")
        raise typer.Exit(code=1) from e

    request_json = json_encode(request_dict)

    logger.debug("Processed request json: %s", request_json)

    try:
        res = json_decode(request_json, type=NewJob)
        if res.request.scheduling.interactive:
            # interactive jobs may not have not_after set
            res.request.scheduling.not_after = None
        logger.debug("Processed NewJob: %s", res)
    except ValidationError as e:
        err_console.print("[red]Errors in request file:")
        err_console.print(f"[red]{str(e)}")
        raise typer.Exit(code=1) from e
    else:
        return res


# TODO: add
# run_interactive



"""AI Infrastructure Service CLI - common code."""

import re
from collections.abc import Collection
from enum import Enum

import typer
from slices_clientlib_aiis.schema.job import Job, JobStatus
from slices_clientlib_core.base.exceptions import BadRequestException, NotFoundException

from . import console, err_console
from .ai_context import AIContext


def version_callback(version: bool):
    """Show CLI version."""
    if version:
        from importlib.metadata import version as metadata_version  # noqa: PLC0415

        console.print(f"v{metadata_version('slices-cli-ai')}")
        raise typer.Exit(0)


def get_requested_ai_jobs(
    ctx: AIContext,
    job_ids: list[str] | None,
    *,
    filter_user: bool = False,
    filter_project: bool = False,
    include_pending: bool = False,
    include_running: bool = False,
    include_finished: bool = False,
    include_deleted: bool = False,
) -> Collection[Job]:
    """Return the requested jobs.

    Be careful, if no job_ids are specified, this returns ALL jobs!
    They are returned as a "PagedCollection", which means they will only fetch jobs as required.
    You can use len() to get the total number of jobs, without requesting all of them.
    """

    allowed_states = []
    if include_pending:
        allowed_states.extend((JobStatus.ONHOLD, JobStatus.QUEUED))
    if include_finished:
        allowed_states.extend(
            (
                JobStatus.FINISHED,
                JobStatus.FAILED,
                JobStatus.CANCELLED,
                JobStatus.HALTED,
            )
        )
    if include_running:
        allowed_states.extend(
            (
                JobStatus.RUNNING,
                JobStatus.STARTING,
                JobStatus.MUSTHALT,
                JobStatus.HALTING,
            )
        )
    if include_deleted:
        allowed_states.append(JobStatus.DELETED)

    if not job_ids:  # all resources
        return (
            ctx.obj.api.for_project(ctx.obj.require_project_id())
            .ai(ctx.obj.site_id)
            .jobs.list(
                user_id=None if not filter_user else ctx.obj.get_logged_in_userinfo("sub"),
                project_id=None if not filter_project else ctx.obj.require_project_id(),
                # Note: allowed_states=None means all except DELETED
                allowed_states=allowed_states,
            )
        )

    return [get_ai_job(ctx, job_id) for job_id in job_ids]


def duration_string_to_seconds(duration: str) -> int | None:  # noqa: PLR0911
    """Convert a duration string (e.g. "1h", "45s, "2d", ...) to seconds."""

    if not duration:
        return None
    match = re.match(r"^([0-9.]+) *([a-zA-Z]+)$", duration)
    if not match:
        return None

    num = float(match.group(1)) if "." in match.group(1) else int(match.group(1))
    unit = match.group(2).lower()
    if unit in ("h", "hour", "hours"):
        return int(num * 60 * 60)
    if unit in ("m", "min", "mins", "minute", "minutes"):
        return int(num * 60)
    if unit in ("s", "sec", "secs", "second", "seconds"):
        return int(num)
    if unit in ("day", "days"):
        return int(num * 60 * 60 * 24)
    if unit in ("week", "weeks"):
        return int(num * 60 * 60 * 24 * 7)
    return None
    # Impossible to support because not fixed length: month, year, ...


def validate_duration_option(
    ctx: typer.Context,
    _param: typer.CallbackParam,
    value: str,
) -> str:
    """Validate a textual timedelta (used for typer arguments)."""

    if ctx.resilient_parsing:
        return value

    seconds = duration_string_to_seconds(value)
    if seconds is None:
        raise typer.BadParameter(f"{value!r} is not a valid duration string (e.g. 1h, 45s, 2d, ...).")

    return value


def status_to_color(status: JobStatus) -> str:  # noqa: PLR0911
    """Return a matching color for a JobStatus."""

    if status == JobStatus.QUEUED:
        return ""
    if status == JobStatus.FAILED:
        return "[red]"
    if status in {JobStatus.STARTING, JobStatus.ASSIGNED}:
        return "[blue]"
    if status == JobStatus.RUNNING:
        return "[dark_blue]"
    if status == JobStatus.FINISHED:
        return "[green]"
    if status in {JobStatus.MUSTHALT, JobStatus.HALTING, JobStatus.HALTED}:
        return "[orange]"
    if status in {JobStatus.CANCELLED, JobStatus.DELETED}:
        return "[grey]"
    return ""


def get_ai_job(ctx: AIContext, job_id: str) -> Job:
    """Get a Job from a full or partial ID."""

    try:
        return ctx.obj.api.for_project(ctx.obj.require_project_id()).ai(ctx.obj.site_id).jobs.get(job_id)
    except NotFoundException:
        err_console.print(f"[red] :x: Job {job_id} not found")
        raise typer.Exit(code=1) from None
    except BadRequestException as e:
        err_console.print(f"[red] :x: Problem fetching job {job_id}: {e.detail}")
        raise typer.Exit(code=1) from None


class ShowResourceFormat(Enum):
    """Show resource format Options."""

    text = "text"
    json = "json"
    yaml = "yaml"


def user_id_to_name(ctx: AIContext, user_id: str) -> str:
    """Convert a user ID to a name, if possible."""

    if user_id == ctx.obj.get_logged_in_userinfo("sub"):
        return ctx.obj.get_logged_in_userinfo("preferred_username")

    try:
        user_info = ctx.obj.api.user.get(user_id)
    except:  # noqa: E722
        return user_id
    else:
        return user_info.preferred_username or user_info.email or user_id


def project_id_to_name(ctx: AIContext, project_id: str) -> str:
    """Convert a project ID to a name, if possible."""

    if project_id == ctx.obj.get_current_project_id():
        return ctx.obj.require_project_friendly_name()

    try:
        project_info = ctx.obj.api.project.get(project_id)
    except:  # noqa: E722
        return project_id
    else:
        return project_info.name or project_id



"""AI Infrastructure Service CLI - Create Job."""

import os
import time
from datetime import datetime, timedelta
from typing import Annotated, Literal, Union
from uuid import UUID

import typer
from pytz import UTC
from rich.progress import Progress
from slices_clientlib_aiis.schema.job import JobStatus

from . import console
from .ai import ai_app
from .ai_context import AIContext
from .ai_helpers import (
    handle_some_exceptions,
)
from .utils import td_format


def _is_active_status(status: JobStatus) -> bool:
    """Return True if the status is active (i.e. the job is still running or pending)."""
    return status in (
        JobStatus.QUEUED,
        JobStatus.ASSIGNED,
        JobStatus.STARTING,
        JobStatus.RUNNING,
        JobStatus.MUSTHALT,
        JobStatus.HALTING,
        JobStatus.CANCELLING,
    )


class SimpleSpinner:
    """A very simple status spinner.

    It creates a spinner by overwriting the same line, using carriage return.
    """

    SIMPLE_PROGRESS_CHARS = ["/", "-", "\\", "|"]

    def __init__(self, msg: str):
        try:
            self.simple_progress_cols = int(os.getenv("COLUMNS", "30"))
        except ValueError:
            self.simple_progress_cols = 30
        self.char_index = 0
        self.msg = msg

    def start(self):
        """Start the spinner."""
        self.print_spinner()

    def print_spinner(self):
        """Print the spinner."""
        ch = self.SIMPLE_PROGRESS_CHARS[self.char_index % len(self.SIMPLE_PROGRESS_CHARS)]
        # console.print(f"{ch: <{self.simple_progress_cols}}", end="\r")
        console.print(f"{f'{ch} {self.msg}': <{self.simple_progress_cols}}", end="\r")
        self.char_index += 1

    def stop(self):
        """Clear the line."""
        console.print(" " * self.simple_progress_cols, end="\r")


def wait_for_job_status(
    ctx: AIContext, job_id: Union[str, UUID], requested_status: JobStatus, max_wait_in_s: int = 60 * 15
):
    """Wait for the job to reach the requested status."""

    start_wait = datetime.now(UTC)
    deadline = start_wait + timedelta(seconds=max_wait_in_s)

    def time_str(show_relative: bool = True):
        if show_relative:
            rel_time = datetime.now(UTC) - start_wait
            rel_time_str = td_format(rel_time)
            needed_spaces = 10 - len(rel_time_str)
            if needed_spaces > 0:
                rel_time_str += " " * needed_spaces
            rel_time_str = f" {rel_time_str} -"
        else:
            rel_time_str = " " * 13
        return f"{datetime.now().astimezone():%Y-%m-%d %H:%M:%S %z} -{rel_time_str}"

    if console.is_dumb_terminal or not console.is_terminal:
        progress = SimpleSpinner(f"[blue]Waiting for job {job_id} to reach status {requested_status.name}...")
        out = console
    else:
        progress = Progress(transient=True)
        progress.start()
        out = progress.console
        progress.add_task(f"[blue]Waiting for job {job_id} to reach status {requested_status.name}...", total=None)
    try:
        last_status = None
        while True:
            if datetime.now(UTC) > deadline:
                out.print(
                    f"[bold red]:x: Waited for job {job_id} to reach status {requested_status.name} "
                    f"for more than {max_wait_in_s} seconds, giving up."
                )
                raise typer.Exit(1)
            job = ctx.obj.api.for_project(ctx.obj.require_project_id()).ai(ctx.obj.site_id).jobs.get(job_id)
            cur_status = job.state.status

            if job.id != job_id:
                # if job_id was a partial ID, it is replaced with a full job UUID
                # if job_id was a str, it is replaced with a UUID
                job_id = job.id

            if cur_status == requested_status:
                out.print(f"[bold green]:sparkles:Job {job_id} reached status {requested_status.name}.")
                return

            if not _is_active_status(cur_status):
                out.print(
                    f"[bold red]:x: Job {job_id} reached final status {cur_status.name}, "
                    f"which is not the requested status {requested_status}."
                )
                raise typer.Exit(1)

            if last_status is None:
                out.print(f"{time_str(show_relative=False)} Job status: {cur_status.name}")
            elif last_status != cur_status:
                out.print(f"{time_str()} Job changed status to: {cur_status.name}")

            if isinstance(progress, SimpleSpinner):
                progress.print_spinner()
            time.sleep(2)
            last_status = cur_status
    finally:
        if progress:
            progress.stop()


@ai_app.command(rich_help_panel="Auxiliary commands")
@handle_some_exceptions
def wait(
    ctx: AIContext,
    job_id: Annotated[str, typer.Argument(help="job ID", show_default=False)],
    target_status: Annotated[
        Literal["RUNNING", "FINISHED"], typer.Argument(help="Target job status to wait for.", case_sensitive=False)
    ] = "FINISHED",
):
    """Wait until the requested status has been reached (or can never be reached)."""
    target_status_value = JobStatus(target_status)
    wait_for_job_status(ctx, job_id, target_status_value)



"""Slices AI CLI Entrypoint."""

import json
import logging
import os
import sys
from pathlib import Path
from typing import Annotated

import msgspec
import typer
from msgspec import ValidationError
from slices_clientlib_aiis.base.config import SlicesAIApiClientsConfig
from slices_clientlib_core.base.auth import (
    CredentialsBrokenError,
    load_auth_token_and_user,
)
from slices_clientlib_core.base.config import (
    ConfigBrokenError,
    load_cli_central_api_clients_config,
    load_current_project_id,
)

from slices_cli_ai.ai_common import version_callback
from slices_cli_ai.util.typer import AliasGroup

from . import VERBOSITY_DEBUG, VERBOSITY_TRACE, console, err_console
from .ai import ai_app
from .api import ApiClients
from .context import Context, ContextObj

# A hack to make help etc show "slices" instead of "slices-ai"
# Except when using shell completion or installing completions
if (
    sys.argv[0].endswith("-ai")
    and "_SLICES_AI_COMPLETE" not in os.environ
    and not ("--install-completion" in sys.argv or "--show-completion" in sys.argv)
):
    sys.argv[0] = sys.argv[0][: -len("-ai")]


logging.basicConfig(level=logging.WARNING)

CONTEXT_SETTINGS = {
    "help_option_names": ["-h", "--help"],
}

app = typer.Typer(
    help="Slices CLI ai",
    context_settings=CONTEXT_SETTINGS,
    no_args_is_help=True,
    cls=AliasGroup,
)
# AliasGroup.register_hidden_alias("foobar", "foobars")


def manifest_callback(manifest: bool):
    """Show slices CLI manifest."""
    if manifest:
        print(
            json.dumps(
                {
                    # The @type that identifies this as a slices cli subcommand manifest
                    "@type": "slices-cli:subcommand-metadata",
                    # The version of the slices cli subcommand manifest (1.0 is the currently the only version)
                    "version": "1.0",
                    #
                    # Whether this subcommand supports command line tab completion.
                    "tab_completion_supported": True,
                    "tab_completion_install_option": "--install-completion",
                    "tab_completion_show_option": "--show-completion",
                    # The help text for this subcommand, to be shown in the main slices CLI help.
                    "short_help": "AI Infrastructure Service",
                }
            )
        )
        raise typer.Exit(0)


@app.callback(invoke_without_command=True)
def initialize_api_clients(
    ctx: Context,
    ai_api_config: Annotated[
        Path | None, typer.Option(exists=True, dir_okay=False, hidden=True, envvar="SLICES_AI_CUSTOM_CONFIG")
    ] = None,
    _version: Annotated[
        bool | None,
        typer.Option(
            "--version",
            callback=version_callback,
            help="Show CLI version.",
            is_eager=True,
        ),
    ] = None,
    verbosity: Annotated[
        int,
        typer.Option(
            "-v",
            help="Verbosity level.",
            count=True,
        ),
    ] = 0,
    _manifest: Annotated[
        bool | None,
        typer.Option(
            "--cli-manifest",
            callback=manifest_callback,
            help="Show Slices CLI manifest.",
            is_eager=True,
            hidden=True,
        ),
    ] = None,
    api_config: Annotated[
        Path | None, typer.Option(exists=True, dir_okay=False, hidden=True, envvar="SLICES_CUSTOM_CONFIG")
    ] = None,
):
    """Initialize API clients."""

    # First we get the "central config". SlicesClient needs this.
    # This will get automatically get the config from the command line option, the env var, the cache, or online source.
    try:
        central_config = load_cli_central_api_clients_config(api_config)
        if verbosity >= VERBOSITY_TRACE:
            console.print(f"[bright_black]Loaded central config: {central_config!r}")
    except ConfigBrokenError as e:
        err_console.print(f"[bold red]{e} :x:")
        raise typer.Exit(1) from None

    # Next, we'll get the auth token and user info.
    # This will automatically load the auth token from the cache, and refresh it if needed.
    # This also automatically supports external auth token providers that use SLICES_AUTH_TOKEN_PROVIDER.
    try:
        auth_token, user_info = load_auth_token_and_user()
        if verbosity >= VERBOSITY_TRACE:
            console.print(f"[bright_black]Loaded auth token: {auth_token}")
        if verbosity >= VERBOSITY_DEBUG:
            console.print(f"[bright_black]Authenticated User ID: {user_info['sub'] if user_info else None}")
    except CredentialsBrokenError as e:
        err_console.print(f"[bold red]{e} :x:")
        err_console.print("[red]Please log in again")
        user_info = None
        raise typer.Exit(1) from None

    # We choose to not support the case where the user is not logged in.
    if user_info is None:
        err_console.print("[bold red]You must be logged in to use this command :x:")
        raise typer.Exit(1)

    # We get the current project ID.
    # This automatically loads the project ID from the setting, or from the env var.
    project_id = load_current_project_id()
    if verbosity >= VERBOSITY_DEBUG:
        console.print(f"[bright_black]Loaded project ID: {project_id!r}")

    ai_config = None
    if ai_api_config:
        try:
            with ai_api_config.open() as f:
                config_str = f.read()
            config_dict = json.loads(config_str)
            if config_dict.get("pubkey") or config_dict.get("bi"):
                err_console.print("[bold red]You are using an old version of the API config. Please upgrade. :x:")
                raise typer.Exit(1)

            if config_dict.get("bi_infrastructures"):
                ai_config = msgspec.json.decode(config_str, type=SlicesAIApiClientsConfig)

            if not ai_config:
                err_console.print(
                    f"[bold red]Your AI config file ('{ai_api_config}') does not contain a valid AI config:x:"
                )
                raise typer.Exit(1)
        except ValidationError as err:
            err_console.print("[bold red]Something is wrong with your AI api configuration :x:")
            err_console.print("\n", str(err))
            raise typer.Exit(1) from err

    api_clients = ApiClients(central_config=central_config, ai_config=ai_config, auth_token=auth_token)
    ctx.obj = ContextObj(api=api_clients, user_info=user_info, project_id=project_id, verbosity=verbosity)


app.add_typer(ai_app, name="ai", rich_help_panel="Commands with subcommands")


if __name__ == "__main__":
    try:
        app()
    except ValidationError as err:
        err_console.print("[bold red]Something is wrong with your input :x:")
        err_console.print(f"[red]{err}[/]")
        raise typer.Exit(1) from err



"""CLI configuration management."""

import logging

from slices_clientlib_core.base.auth import load_auth_token as core_load_auth_token
from slices_clientlib_core.base.exceptions import ConfigurationError
from slices_clientlib_core.oauth2.token import OAuthToken

from . import err_console

logger = logging.getLogger(__name__)


def load_auth_token() -> OAuthToken | None:
    """Load authentication token from file."""
    try:
        return core_load_auth_token()
    except ConfigurationError as err:
        err_console.print("[bold red]Could not load auth token from external provider :x:")
        err_console.print(f"[red]{err}")



"""Helpers for CLI argument str to timedelta."""

import re
from datetime import datetime, timedelta

import typer
from pytz import UTC

STR_TO_TIMEDELTA_PATTERN = re.compile(r"^([0-9]+) *([a-z]+)$")


STR_TO_TIMEDELTA_UNITS = [
    "s",
    "second",
    "seconds",
    "min",
    "minute",
    "minutes",
    "h",
    "hour",
    "hours",
    "d",
    "day",
    "days",
    "w",
    "week",
    "weeks",
]
STR_TO_TIMEDELTA_HELP = f"'<number>[optional space]<unit>' (where unit one of: {', '.join(STR_TO_TIMEDELTA_UNITS)})"


def str_to_timedelta(duration: str) -> timedelta:
    """Convert a str to a timedelta object.

    This assumes <count:int>[<space>]<unit:str>
    Where unit is one of: s/second/seconds min/minute/minutes h/hour/hours d/day/days w/week/weeks

    Raises ValueError for non convertible str.
    """
    match = STR_TO_TIMEDELTA_PATTERN.match(duration.strip())
    if not match:
        raise ValueError(f"Invalid duration: {duration}")
    count = int(match.group(1))
    unit = match.group(2)
    if unit in ["s", "second", "seconds"]:
        return timedelta(seconds=count)
    if unit in ["m", "minute", "minutes"]:
        return timedelta(minutes=count)
    if unit in ["h", "hour", "hours"]:
        return timedelta(hours=count)
    if unit in ["d", "day", "days"]:
        return timedelta(days=count)
    if unit in ["w", "week", "weeks"]:
        return timedelta(weeks=count)
    raise ValueError(f"Invalid duration unit: {duration}")


def validate_timedelta_str(ctx: typer.Context, param: typer.CallbackParam, value: str) -> str:  # noqa: ARG001
    """Validate a textual timedelta (used for typer arguments)."""
    try:
        str_to_timedelta(value)
    except ValueError as e:
        # raise typer.BadParameter(f"Invalid --{param.name} value {value!r}") from e
        raise typer.BadParameter(
            f"{value!r} is not a valid duration. A duration must be of the format: {STR_TO_TIMEDELTA_HELP}"
        ) from e
    return value


DATETIME_FORMATS = [
    "%Y-%m-%d",
    "%Y-%m-%dT%H:%M",
    "%Y-%m-%d %H:%M",
]


def validate_datetime_str(value: str) -> datetime:  # noqa: ARG001
    """Validate a textual datetime (used for typer arguments)."""
    for datetime_format in DATETIME_FORMATS:
        try:
            return datetime.strptime(value, datetime_format).astimezone()
        except ValueError:  # noqa: PERF203
            pass
    raise typer.BadParameter(f"{value!r} does not match any of the formats: {', '.join(DATETIME_FORMATS)}")


def duration_or_datetime(value: str) -> datetime:  # noqa: ARG001
    """Parse a textual timedelta or datetime (used for typer arguments).

    When a timedelta is detected, the result is the current datetime plus the timedelta.
    """

    try:
        td = str_to_timedelta(value)
        return datetime.now(UTC).replace(second=0, microsecond=0) + td
    except ValueError:
        pass

    try:
        return validate_datetime_str(value)
    except typer.BadParameter:
        pass

    raise typer.BadParameter(
        f"{value!r} is not a date, nor a duration.\n"
        f"A date must match any of the formats: {', '.join(DATETIME_FORMATS)}.\n"
        f"A duration must be of the format: {STR_TO_TIMEDELTA_HELP}"
    )



"""Public key utilities."""

import base64
import binascii
import hashlib


def sha256_fingerprint(public_key: str) -> str:
    """SHA256 fingerprint of the public key."""
    data = public_key.split(" ")[1].encode()

    digest = hashlib.sha256(binascii.a2b_base64(data)).digest()
    encoded = base64.b64encode(digest).rstrip(b"=")  # ssh-keygen strips these
    return "SHA256:" + encoded.decode()



"""AI Infrastructure Service CLI - action commands."""

from datetime import datetime, timedelta
from typing import Annotated

import typer
from pytz import UTC
from slices_clientlib_core.base.exceptions import ConflictException

from . import console, err_console
from .ai import ai_app
from .ai_common import duration_string_to_seconds, get_ai_job, validate_duration_option
from .ai_context import AIContext
from .ai_helpers import (
    handle_some_exceptions,
)
from .ai_subcommands_create import get_job_max_end_time_s


@ai_app.command("cancel", rich_help_panel="Lifecycle commands")
@handle_some_exceptions
def cancel_job(
    ctx: AIContext,
    job_id: Annotated[str, typer.Argument(help="Job ID.", show_default=False)],
):
    """Cancel a job."""

    project_id = ctx.obj.get_current_project_id()
    if not project_id:
        raise typer.Exit(0)

    job = get_ai_job(ctx, job_id)

    ctx.obj.api.for_project(ctx.obj.require_project_id()).ai(ctx.obj.site_id).jobs.cancel(job.id)

    console.print(f"[green] :recycle: Cancelled Job {job.id}")


@ai_app.command("rm", rich_help_panel="Lifecycle commands")
@handle_some_exceptions
def delete_job(
    ctx: AIContext,
    job_id: Annotated[str, typer.Argument(help="Job ID.", show_default=False)],
):
    """Delete a job."""

    project_id = ctx.obj.get_current_project_id()
    if not project_id:
        raise typer.Exit(0)

    job = get_ai_job(ctx, job_id)

    ctx.obj.api.for_project(ctx.obj.require_project_id()).ai(ctx.obj.site_id).jobs.delete(job.id)

    console.print(f"[green] :recycle: Deleted Job {job.id}")


@ai_app.command("halt", rich_help_panel="Lifecycle commands")
@handle_some_exceptions
def halt_job(
    ctx: AIContext,
    job_id: Annotated[str, typer.Argument(help="Job ID.", show_default=False)],
):
    """Halt a job. Halted jobs may be automatically re-QUEUED later."""

    project_id = ctx.obj.get_current_project_id()
    if not project_id:
        raise typer.Exit(0)

    job = get_ai_job(ctx, job_id)

    ctx.obj.api.for_project(ctx.obj.require_project_id()).ai(ctx.obj.site_id).jobs.halt(job.id)

    console.print(f"[red] :stop_sign: [green] Requested Job halt for {job.id}")


@ai_app.command("hold", rich_help_panel="Lifecycle commands")
@handle_some_exceptions
def hold_job(
    ctx: AIContext,
    job_id: Annotated[str, typer.Argument(help="Job ID.", show_default=False)],
):
    """Hold a QUEUED job, preventing it from running."""

    project_id = ctx.obj.get_current_project_id()
    if not project_id:
        raise typer.Exit(0)

    job = get_ai_job(ctx, job_id)

    ctx.obj.api.for_project(ctx.obj.require_project_id()).ai(ctx.obj.site_id).jobs.hold(job.id)

    console.print(
        f"[green] :pause_button: Holding Job {job.id} (It will not run until you release it to the queue again.)"
    )


@ai_app.command("requeue", rich_help_panel="Lifecycle commands")
@handle_some_exceptions
def requeue_job(
    ctx: AIContext,
    job_id: Annotated[str, typer.Argument(help="Job ID.", show_default=False)],
):
    """Requeue/Release a held job, QUEUEing it, so it waits to run."""

    project_id = ctx.obj.get_current_project_id()
    if not project_id:
        raise typer.Exit(0)

    job = get_ai_job(ctx, job_id)

    ctx.obj.api.for_project(ctx.obj.require_project_id()).ai(ctx.obj.site_id).jobs.requeue(job.id)

    console.print(
        f"[green] :play_button: Released Job {job.id} to the queue. (Where it will wait for its turn to run.)"
    )


def validate_duration_or_date_option(ctx: typer.Context, _param: typer.CallbackParam, value: str) -> datetime | None:
    """Validate that the value is either a duration (e.g. 1h30m) or a date (e.g. 2024-12-31T23:59:59Z).

    This always returns a date.
    If the input is a duration, it returns the current time + duration.
    If the input is a date, it returns the date.
    """

    if value is None:
        return None

    if ctx.resilient_parsing:
        return None

    try:
        # if it's a valid duration, it is OK.
        validate_duration_option(ctx, _param, value)
        s: int = duration_string_to_seconds(value)  # pyright: ignore [reportAssignmentType] cannot be None
        return datetime.now(UTC) + timedelta(seconds=s)
    except ValueError:
        pass

    try:
        return datetime.fromisoformat(value.replace("Z", "+00:00"))
    except ValueError:
        raise ValueError(
            f"Invalid value '{value}'. "
            f"Must be either a duration (e.g. 1h, 30m, ...) or a date (e.g. 2024-12-31T23:59:59Z)."
        ) from None


def validate_duration_option_or_none(ctx: typer.Context, param: typer.CallbackParam, value: str | None) -> str | None:
    """Validate that the value is either None or a duration (e.g. 1h, 30m, ...)."""

    if value is None:
        return None

    return validate_duration_option(ctx, param, value)


@ai_app.command("modify", rich_help_panel="Auxiliary commands")
@handle_some_exceptions
def modify_job(
    ctx: AIContext,
    job_id: Annotated[str, typer.Argument(help="Job ID.", show_default=False)],
    min_duration: Annotated[
        str | None,
        typer.Option("--min-duration", help="New minDuration for the job.", callback=validate_duration_option_or_none),
    ] = None,
    max_duration: Annotated[
        str | None,
        typer.Option("--max-duration", help="New maxDuration for the job.", callback=validate_duration_option_or_none),
    ] = None,
    not_after: Annotated[
        datetime | None,
        typer.Option("--not-after", help="New notAfter for the job.", callback=validate_duration_or_date_option),
    ] = None,
):
    """Change maxDuration, minDuration or notAfter of a job."""
    project_id = ctx.obj.require_project_id()

    if min_duration is None and max_duration is None and not_after is None:
        err_console.print(
            "[red] No changes specified. Use --min-duration, --max-duration or --not-after to specify changes."
        )
        raise typer.Exit(1)

    job = get_ai_job(ctx, job_id)

    experiment_id = job.owner.experiment_id
    if experiment_id is None:
        err_console.print(
            f"[red] :x: Job {job_id} does not have an experiment_id, cannot modify scheduling parameters."
        )
        raise typer.Exit(1)

    modify_job_end = max_duration is not None or not_after is not None

    required_expires_at = datetime.now(UTC) + timedelta(
        seconds=get_job_max_end_time_s(
            max_duration or job.request.scheduling.max_duration, not_after or job.request.scheduling.not_after
        )
    )

    experiment = ctx.obj.api.experiment.get(experiment_id)
    if modify_job_end and experiment.expires_at < required_expires_at:
        # Extend expiration time if the existing experiment expires before required expire time
        ctx.obj.api.experiment.update(
            experiment_id,
            project_id,
            expires_at=required_expires_at,
            description=experiment.description,
        )
        console.print("[green] :clock10: Extended experiment.")

    try:
        if max_duration is not None:
            ctx.obj.api.for_project(project_id).for_experiment(experiment_id).ai(
                ctx.obj.site_id
            ).jobs.update_max_duration(job_id, max_duration)
            console.print("[green] :pencil: Modified maxDuration.")

        if min_duration is not None:
            ctx.obj.api.for_project(project_id).for_experiment(experiment_id).ai(
                ctx.obj.site_id
            ).jobs.update_min_duration(job_id, min_duration)
            console.print("[green] :pencil: Modified minDuration.")

        if not_after is not None:
            ctx.obj.api.for_project(project_id).for_experiment(experiment_id).ai(ctx.obj.site_id).jobs.update_not_after(
                job_id, not_after
            )
            console.print("[green] :pencil: Modified notAfter.")
    except ConflictException as e:
        err_console.print(f"[red] :x: Could not modify job: {e}")
        raise typer.Exit(1) from None



"""Slices AI CLI authentication."""

import os
from typing import Annotated

import typer
from requests.exceptions import HTTPError
from slices_clientlib_core.base.auth import auth_file_location, delete_auth_token, parse_auth_token, save_auth_token
from slices_clientlib_core.base.config import (
    set_config_string,
)
from slices_clientlib_core.oauth2.flow import AuthenticationFlow, DeviceFlow

from . import console, err_console
from .context import Context

auth_app = typer.Typer(help="User authentication", no_args_is_help=True)


@auth_app.command(help="Authenticate and retrieve credentials", rich_help_panel="Commands")
def login(
    ctx: Context,
    open_browser: Annotated[bool, typer.Option(help="Automatically open browser to perform login.")] = False,
):
    """Authenticate and retrieve credentials."""

    if os.environ.get("SLICES_AUTH_TOKEN_PROVIDER"):
        err_console.print("[bold red]Authentication is handled by external token provider.")

    if not open_browser:
        flow = DeviceFlow()

        console.print(f"Please go to {flow._device_code_info['verification_uri_complete']} and login to continue.")
        try:
            with console.status("[mangenta]Waiting for authorization"):
                token = flow.wait_for_authorization_token()
        except TimeoutError:
            err_console.print("The user code has expired. Please start the login flow again.")
            raise typer.Exit(1) from None
        except ValueError:
            err_console.print("Could not retrieve authentication details from device login flow.")
            raise typer.Exit(1) from None
    else:
        flow = AuthenticationFlow()

        try:
            with console.status("[mangenta]Waiting for authentication response"):
                token = flow.wait_for_authorization_token()
        except TimeoutError:
            err_console.print("Timeout while waiting for authorization token.")
            raise typer.Exit(1) from None
        except ValueError:
            err_console.print("Could not exchange authorization code for a token.")
            raise typer.Exit(1) from None

    if token:
        save_auth_token(token)

        auth_info = parse_auth_token(token)
        console.print(f"[bold green]Successfully logged in as '{auth_info['preferred_username']}' ")
        console.print(f":floppy_disk: Saved to {auth_file_location}")

        if ctx.obj.get_current_project_id() is None:
            console.print(
                ":backhand_index_pointing_right: "
                "Now select the project that you want to use with [blue]slices project use <project name>[/]"
            )

    else:
        err_console.print("[red]Logging in failed. :x:")
        raise typer.Exit(1)


@auth_app.command(help="Show user credentials.", rich_help_panel="Commands")
def show(  # noqa: PLR0912
    ctx: Context,
    show_id: Annotated[bool, typer.Option("--show-id", "-I", help="Show internal ID.")] = False,
    show_access_token: Annotated[bool, typer.Option(help="Show JWT Access Token.")] = False,
    show_id_token: Annotated[bool, typer.Option(help="Show JWT ID Token.")] = False,
    show_refresh_token: Annotated[bool, typer.Option(help="Show JWT Refresh Token.")] = False,
):
    """Show user credentials."""
    user_info = ctx.obj.user_info

    if user_info:
        if not show_id:
            console.print(f"Logged in as '{user_info['preferred_username']}'")
        else:
            console.print(f"Logged in as '{user_info['sub']}'")

        if show_access_token:
            if ctx.obj.api.auth_token:
                console.print(f"Access Token: '{ctx.obj.api.auth_token.access_token}'")
            else:
                console.print("Access Token: none")
        if show_id_token:
            if ctx.obj.api.auth_token:
                console.print(f"ID Token: '{ctx.obj.api.auth_token.id_token}'")
            else:
                console.print("ID Token: none")
        if show_refresh_token:
            if ctx.obj.api.auth_token:
                console.print(f"Refresh Token: '{ctx.obj.api.auth_token.refresh_token}'")
            else:
                console.print("Refresh Token: none")
    else:
        err_console.print("[red]Not logged in. :x:")
        raise typer.Exit(1)


@auth_app.command(help="Remove user credentials", rich_help_panel="Commands")
def logout():
    """Remove user credentials."""
    if os.environ.get("SLICES_AUTH_TOKEN_PROVIDER"):
        err_console.print("[red]Cannot logout when using an external token provider. :x:")
        raise typer.Exit(1)

    delete_auth_token()
    set_config_string("default", None, "project")
    console.print("[bold green]Successfully logged out.[/] :white_check_mark:")
    console.print(f":floppy_disk: Deleted {auth_file_location}")


@auth_app.command(help="Get credentials for a specific audience.", rich_help_panel="Advanced Commands")
def get_for_audience(
    ctx: Context,
    audience: str,
    with_project: Annotated[bool, typer.Option(help="Include project ID in the token.")] = True,
):
    """Print user credentials for a specific audience."""
    user_info = ctx.obj.user_info

    if not user_info:
        err_console.print("[red]Not logged in. :x:")
        raise typer.Exit(1)
    try:
        print(
            ctx.obj.api.get_access_token_for_audience(audience, ctx.obj.require_project_id() if with_project else None)
        )
    except HTTPError as e:
        try:
            if e.response:
                error_response = e.response.json()
                err_console.print(
                    f"[bold red]Error: Could not get token. :x:[/]\n[red]{error_response['error_description']}"
                )
            else:
                err_console.print("[bold red]Error: Could not get token. :x:")
        except:  # noqa: E722
            err_console.print("[bold red]Error: Could not get token. :x:")
            if e.response:
                err_console.print(f"[red]{str(e.response.text)}")
        raise typer.Exit(1) from None


@auth_app.command(help="Get id_token for a specific audience.", rich_help_panel="Advanced Commands")
def id_token(
    ctx: Context,
    audience: str,
):
    """Print id_token for a specific audience."""

    print(ctx.obj.api.get_id_token_for_audience(audience, ctx.obj.require_project_id()))



"""CLI Utils."""

import json
import re
import shlex
from collections.abc import Collection
from datetime import datetime, timedelta
from json import JSONDecodeError
from pathlib import Path
from typing import Any

import typer
from slices_clientlib_core.base.exceptions import (
    ApiException,
    BadRequestException,
    NotFoundException,
    ServiceException,
    UnauthorizedException,
    ValidationError,
)
from slices_clientlib_core.pubkey import PublicKey

PUBLIC_SSH_KEY_LOCATIONS = ["~/.ssh/id_ed25519.pub", "~/.ssh/id_ecdsa.pub", "~/.ssh/id_rsa.pub"]


def load_public_ssh_keys() -> list[str]:
    """Load SSH keys from well known locations."""
    public_ssh_keys = []
    for ssh_key_location in PUBLIC_SSH_KEY_LOCATIONS:
        path = Path(ssh_key_location).expanduser()
        if path.exists():
            public_ssh_keys.append(path.read_text())

    return public_ssh_keys


local_timezone = datetime.now().astimezone().tzinfo


def shlex_quote_if_needed(arg: str) -> str:
    """Quote for shell, if needed."""
    res = shlex.quote(arg)
    if res == f"'{res}'":
        # simple quoting
        return arg
    return res


def create_add_ssh_pubkey_command(
    base_ssh_command: list[str], pubkeys_to_add: Collection[PublicKey], member_name: str | None
) -> str:
    """Create a command for accessing a node via SSH."""
    pubkey_lines = [f"COMMENT='{member_name or ''}'"]
    for pubkey in pubkeys_to_add:
        pubkey_lines.append(f"PUBKEY='{pubkey.public_key}'")
        pubkey_lines.append(
            'grep -qsF -- "$PUBKEY" ~/.ssh/authorized_keys || echo "$PUBKEY $COMMENT" >> ~/.ssh/authorized_keys'
        )

    escaped_ssh_command = " ".join([shlex_quote_if_needed(a) for a in base_ssh_command])

    lines = "\n".join(pubkey_lines)
    return f"""{escaped_ssh_command} 'sh -s' <<'EOF'\n{lines}\nexit 0\nEOF\n"""


def friendly_datetime(dt: datetime, *, include_seconds: bool = False):
    """Format datetime in local timezone in a human-readable manner."""
    # return dt.astimezone(local_timezone).strftime("%Y-%m-%d %H:%M")
    return dt.astimezone(local_timezone).strftime(
        "%Y-%m-%d %H:%M %Z" if not include_seconds else "%Y-%m-%d %H:%M:%S %Z"
    )


def datetime_with_local_timezone(dt: datetime):
    """Make naive datetime into a datetime with local timezone."""
    if dt.tzinfo is not None:
        raise ValueError("Expected naive datetime object.")
    return dt.astimezone(local_timezone)


def get_error_detail(error_body: Any | None):
    """Get detail from error response."""
    if not isinstance(error_body, str):
        return None
    try:
        error_body_contents = json.loads(error_body)
    except JSONDecodeError:
        return error_body

    error_details = []

    if "detail" in error_body_contents:
        if isinstance(error_body_contents["detail"], list):
            for detail in error_body_contents["detail"]:
                if "msg" in detail:
                    error_details.append(detail["msg"])
                else:
                    error_details.append(str(detail))
        else:
            error_details.append(error_body_contents["detail"])

    if error_details:
        res = []
        for detail in error_details:
            if isinstance(detail, str):
                res.append(detail)
            elif isinstance(detail, dict | list):
                res.append(json.dumps(detail, indent=4))
            else:
                res.append(str(detail))
        return "\n".join(res)

    return "No details available"


def get_api_error_detail(err: ApiException):
    """Show error details for API exceptions."""

    if isinstance(err, ValidationError):
        return "\n".join([f"{error_detail.loc}: {error_detail.msg}" for error_detail in err.details])

    if isinstance(err, BadRequestException):
        return f"Bad Request: {err.detail}"

    if isinstance(err, ServiceException):
        message = "The server encountered an internal error"
        if err._httpx_error:
            message += (
                f" while processing {err._httpx_error.request.method} request "
                f"on {err._httpx_error.request.url}: {err._httpx_status_error.response.text}"
            )
        else:
            message += "."
        message += "\n\nPlease report this issue to support if the issue persists."

        return message

    if isinstance(err, UnauthorizedException):
        return f"Unauthorized: {err._httpx_error}"

    if isinstance(err, NotFoundException):
        assert err._httpx_error
        return f"Resource not found: {err._httpx_error.request.url}"

    return f"TODO: Implement get_api_error_detail for other error types {err} {err!r}"


def is_type_id(s: str, type: str | None = None, domain: str | None = None):
    """Is given string a valid TypeID."""
    type_pattern = type if type else r"[a-z]{1,6}"
    domain_pattern = domain if domain else r"[a-z0-9-.]{2,56}"
    suffix_pattern = r"[01234567][0123456789abcdefghjkmnpqrstvwxyz]{25}"
    pattern = re.compile(f"^{type_pattern}_{domain_pattern}_{suffix_pattern}$")

    return pattern.match(s)


def is_experiment_id(s: str):
    """Is given string a valid experiment TypeID."""
    return is_type_id(s, type="exp")


def is_project_id(s: str):
    """Is given string a valid project TypeID."""
    return is_type_id(s, type="proj")


def validate_type_id(value: str, *, required_type: str | None = None):
    """Validate TypeID."""

    if value is None:
        return None

    if not is_type_id(value):
        raise typer.BadParameter("invalid ID")

    type_, domain, suffix = value.split("_")
    if type_ is not None and type_ != required_type:
        raise typer.BadParameter(f"invalid ID type. Expected {required_type!r}, got {type_!r}")

    return value


def td_format(td_object: timedelta):
    """Generate user-friendly time duration."""

    seconds = int(td_object.total_seconds())

    if seconds == 0:
        return "less than 1 second"

    periods = [
        ("year", 60 * 60 * 24 * 365),
        ("month", 60 * 60 * 24 * 30),
        ("day", 60 * 60 * 24),
        ("hour", 60 * 60),
        ("minute", 60),
        ("second", 1),
    ]
    strings = []
    for period_name, period_seconds in periods:
        if seconds >= period_seconds:
            period_value, seconds = divmod(seconds, period_seconds)
            if period_value == 1:
                strings.append(f"{period_value} {period_name}")
            else:
                strings.append(f"{period_value} {period_name}s")

    return ", ".join(strings)



"""AI Infrastructure Service CLI - base typer app."""

import importlib
from pathlib import Path
from typing import Annotated

import typer
from slices_clientlib_aiis import SlicesClientWithAIInfraService

from . import err_console
from .ai_context import AIContextObj
from .context import Context
from .util.typer import AliasGroup

ai_app = typer.Typer(short_help="Basic Infrastructure", no_args_is_help=True, cls=AliasGroup)
AliasGroup.register_hidden_alias("clusters", "cluster")
AliasGroup.register_hidden_alias("list", "jobs")
AliasGroup.register_hidden_alias("show", "job")


def ai_site_callback(ctx: Context, _param: typer.CallbackParam, value: str):
    """Validate bi_infra value."""
    if ctx.resilient_parsing:
        # we are in shell completion, so return anything
        return list(ctx.obj.api.ai_config.ai_service_sites.keys())[0]

    if value not in ctx.obj.api.ai_config.ai_service_sites:
        raise typer.BadParameter(f"Invalid AI Infrastructure Service site ID {value!r}")

    return value


def complete_site_id(incomplete_value: str):
    """Shell completion for infra ID parameter."""
    # Note: this ignores custom config (but we don't have ctx.obj here)
    slices_client = SlicesClientWithAIInfraService(central_config=None, ai_config=None)
    infras_ids = [k for k, v in slices_client.ai_api_clients_config.ai_service_sites.items() if "dev" not in v.tags]
    for infras_id in infras_ids:
        if infras_id.startswith(incomplete_value):
            yield infras_id


DEFAULT_AI_SITE_ID = "be-gent1"


@ai_app.callback(invoke_without_command=False)
def initialize_ai(
    ctx: Context,
    site: Annotated[
        str,
        typer.Option(
            "--site",
            "--site-id",
            envvar=["SLICES_AI_SITE", "SLICES_AI_SITE_ID"],
            callback=ai_site_callback,
            help="AI Infrastructure Service site ID.",
            autocompletion=complete_site_id,
        ),
    ] = DEFAULT_AI_SITE_ID,
):
    """AI Infrastructure Service Commands."""

    if ctx.obj.user_info is None:
        err_console.print("[red]You must be logged in. :x:")
        raise typer.Exit(code=1)

    ctx.obj = AIContextObj(
        api=ctx.obj.api,
        user_info=ctx.obj.user_info,
        site_id=site,
        verbosity=ctx.obj.verbosity,
        project_id=ctx.obj.project_id,
    )


# TODO: add
# print_bugreport_context


def _import_subcommands():
    """Import all ai_subcommands_* files, so we register all the subcommands with typer."""
    current_dir = Path(__file__).parent
    module_files = Path(current_dir).glob("ai_subcommands_*.py")

    for file_path in module_files:
        module_name = file_path.name.replace(".py", "")
        importlib.import_module(f"slices_cli_ai.{module_name}")


_import_subcommands()



"""AI Infrastructure Service CLI - Public SSH Key helpers."""

import typer
from typer._click.types import ParamType

from . import err_console
from .ai_context import AIContext


def get_user_pubkeys(ctx: AIContext) -> list[str]:
    """Get public keys of all members in the experiment."""

    return [
        f"{pubkey.public_key} {ctx.obj.get_logged_in_userinfo('sub')}"
        for pubkey in ctx.obj.api.public_key.list_for_user(ctx.obj.get_logged_in_userinfo("sub"))
    ]


SSH_PUBKEY_PREFIXES = ["ecdsa-", "ssh-", "sk-", "rsa-"]


class SshKeyParamType(ParamType):
    """Custom SSH Key ParamType."""

    envvar_list_splitter = ","
    name = "ssh-key"

    def convert(self, value: str, param, ctx):  # noqa: ARG002
        """Convert SSH key."""
        if not any(value.startswith(prefix) for prefix in SSH_PUBKEY_PREFIXES):
            raise typer.BadParameter("Not a valid SSH public key (in OpenSSH format).")
        return value


def validate_ssh_key(pubkey: str) -> str:
    """Validate SSH public key."""
    if not any(pubkey.startswith(prefix) for prefix in SSH_PUBKEY_PREFIXES):
        raise typer.BadParameter("Not a valid SSH public key (in OpenSSH format).")
    return pubkey


def get_ssh_keys(
    ctx: AIContext, extra_ssh_keys: list[str] | None, extra_ssh_key_files: list[typer.FileText] | None
) -> list[str]:
    """Retrieve all ssh keys that need to be passed to the AI service.

    Shows an error if none can be found.
    """
    ssh_authorized_keys = get_user_pubkeys(ctx)
    if extra_ssh_keys:
        ssh_authorized_keys.extend(extra_ssh_keys)
    if extra_ssh_key_files:
        for file in extra_ssh_key_files or []:
            ssh_authorized_keys.append(validate_ssh_key(file.read()))

    if not ssh_authorized_keys:
        err_console.print("[red]You need to have at least one public SSH key registered. :x:")
        err_console.print("Use: [blue]slices pubkey register")
        raise typer.Exit(code=1)

    return ssh_authorized_keys



"""Typer utils."""

import re

from typer.core import TyperGroup


class AliasGroup(TyperGroup):
    """Add support for command aliases to Typer.

    Extends https://github.com/fastapi/typer/issues/132#issuecomment-2417492805
    Hopefully one day obsolete if this gets added: https://github.com/fastapi/typer/issues/1242
    """

    _hidden_aliases = {}

    @classmethod
    def register_hidden_alias(cls, command, hidden_alias: str):  # noqa D102
        """Very ugly way to add support for aliases.

        TODO: replace with something better, or wait until typer has something built in.
        """
        cls._hidden_aliases.setdefault(command, []).append(hidden_alias)

    def _unalias(self, command):
        for cmd, aliases in self._hidden_aliases.items():
            if command in aliases:
                return cmd
        return command

    _CMD_SPLIT_P = re.compile(r" ?[,|] ?")

    def get_command(self, ctx, cmd_name):  # noqa D102
        cmd_name = self._group_cmd_name(self._unalias(cmd_name))
        return super().get_command(ctx, cmd_name)

    def _group_cmd_name(self, default_name):
        for cmd in self.commands.values():
            name = cmd.name
            if name and default_name in self._CMD_SPLIT_P.split(name):
                return name
        return default_name



"""AI Infrastructure Service CLI - SSH."""

import json
import os
from enum import Enum
from typing import Annotated

import typer
from slices_clientlib_aiis.schema.job import Job, SSHInfo

from . import console, err_console
from .ai import ai_app
from .ai_common import get_ai_job
from .ai_context import AIContext
from .ai_helpers import (
    handle_some_exceptions,
    test_ssh_connect,
)


def _unnest_jump_proxies(first_jump_proxy: SSHInfo | None) -> list[SSHInfo]:
    """Return a list of all jump proxies, instead of the nested representation.

    This is easier to use in code.
    The list is also reversed, that is, the jump proxy all others depend on, comes first.
    """

    res = []
    # The loop below should always be finite.
    # But it is always best to have a safety.
    # Because infinite loops cause end user bugs that are hard to debug.
    safety = 5
    jump_proxy = first_jump_proxy
    while jump_proxy:
        res.insert(0, jump_proxy)
        jump_proxy = jump_proxy.jump_proxy
        safety -= 1
        if safety <= 0:
            err_console.print("[red]BUG: Too much loops handling jump proxy :x:")
            err_console.print("[red]Please report this bug!")
            raise typer.Exit(code=1)

    return res


def _make_jump_proxy_cli_args_and_config(unnested_jump_proxies: list[SSHInfo]) -> tuple[list[str], str]:
    ssh_jump_proxy_destinations = []
    for jump_proxy in unnested_jump_proxies:
        port_dest = f":{jump_proxy.port}" if jump_proxy.port != DEFAULT_SSH_PORT else ""
        ssh_jump_proxy_destinations.append(f"{jump_proxy.username}@{jump_proxy.host}{port_dest}")

    return ["-J", ",".join(ssh_jump_proxy_destinations)], f"JumpProxy {','.join(ssh_jump_proxy_destinations)}"


def _get_jump_ssh_arg(jump: SSHInfo) -> list[str]:
    jump_proxies = _unnest_jump_proxies(jump)
    jump_proxy_cli_args, _ = _make_jump_proxy_cli_args_and_config(jump_proxies)
    return jump_proxy_cli_args


def _print_ssh_login(ssh_login: SSHInfo):
    params = []

    if ssh_login.port != DEFAULT_SSH_PORT:
        params.extend(["-p", f"{ssh_login.port}"])
    if ssh_login.jump_proxy:
        params.extend(_get_jump_ssh_arg(ssh_login.jump_proxy))

    return f"ssh{(' ' + ' '.join(params)) if params else ''} {ssh_login.username}@{ssh_login.host}"


class ProxyOption(Enum):
    """Proxy Options."""

    off = "off"
    on = "on"
    auto = "auto"


class ShowOption(Enum):
    """Show options."""

    nothing = "nothing"
    json = "json"
    command = "command"
    ansible = "ansible"
    ssh_config = "ssh_config"
    server_host = "server_host"
    proxy_host = "proxy_host"
    # server_pubkey_openssh = "server_pubkey_openssh"
    # proxy_pubkey_openssh = "proxy_pubkey_openssh"


DEFAULT_SSH_PORT = 22


def _job_to_ansible(job: Job, ssh_login: SSHInfo, proxy: bool):
    """Convert a resource to an Ansible inventory entry."""
    if proxy and ssh_login.jump_proxy:
        jump_proxies = _unnest_jump_proxies(ssh_login.jump_proxy)
        jump_proxy_cli_args, _ = _make_jump_proxy_cli_args_and_config(jump_proxies)
        jump_proxy_cli_args_str = " ".join(jump_proxy_cli_args)
        ansible_ssh_common_args = f" ansible_ssh_common_args='{jump_proxy_cli_args_str}'"
    else:
        ansible_ssh_common_args = ""

    return (
        f"gpulab-{job.short_id} "
        f"ansible_ssh_port={ssh_login.port} "
        f"ansible_ssh_host={ssh_login.host} "
        f"ansible_ssh_user={ssh_login.username} {ansible_ssh_common_args}"
    )


def _ssh_common(
    ctx: AIContext,
    job_id: str,
    proxy: ProxyOption,
    *,
    sftp_mode: bool = False,  # whether this is for sftp
) -> tuple[Job, SSHInfo, list[str], str, str]:
    job = get_ai_job(ctx, job_id)

    if not job.ssh_info:
        err_console.print("The resource has no SSH logins")
        raise typer.Exit(code=1)

    ssh_login = job.ssh_info[0]

    destination = f"{ssh_login.username}@{ssh_login.host}"

    if proxy == ProxyOption.auto:
        can_connect = test_ssh_connect(ssh_login.host, ssh_login.port)
        proxy = ProxyOption.on if can_connect else ProxyOption.off

        # method from legacy GPULab CLI
        # # Try to resolve worker, if it doesn't assume we need the proxy
        # # (reason: IDLab Gent hostnames have public IPv6, but not IPv4.)
        # try:
        #     host_ip = socket.gethostbyname(job.state.resources.slave_host)
        #     proxy = ProxyOption.off
        #
        #     # Also hardcode that uantwerpen needs proxy.
        #     # (reason: uantwerpen nodes are firewalled)
        #     if job.state.resources.slave_host.endswith(".idlab.uantwerpen.be"):
        #         proxy = ProxyOption.on
        # # except socket.gaierror as e:
        # except OSError as e:
        #     proxy = ProxyOption.on

    ssh_command_options: list[str] = []
    ssh_config_jump_proxy = ""

    if ssh_login.port != DEFAULT_SSH_PORT:
        port_option = "-p" if not sftp_mode else "-P"
        ssh_command_options.append(f"{port_option} {ssh_login.port}")

    jump_proxies = []
    if proxy == ProxyOption.on and ssh_login.jump_proxy:
        jump_proxies = _unnest_jump_proxies(ssh_login.jump_proxy)
        jump_proxy_cli_args, ssh_config_jump_proxy = _make_jump_proxy_cli_args_and_config(jump_proxies)
        ssh_command_options.extend(jump_proxy_cli_args)

    return job, ssh_login, ssh_command_options, destination, ssh_config_jump_proxy


def make_ssh_command(
    ctx: AIContext,
    job_id: str,
    proxy: ProxyOption = ProxyOption.auto,
) -> list[str]:
    """Make the command to connect to the ssh."""
    job, ssh_login, ssh_command_options, destination, ssh_config_jump_proxy = _ssh_common(ctx, job_id, proxy)

    return ["ssh", *ssh_command_options, destination]


@ai_app.command(
    "ssh",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
    rich_help_panel="Job interaction commands",
)
@handle_some_exceptions
def ssh_to_job(
    ctx: AIContext,
    job_id: Annotated[str, typer.Argument(help="Job ID.", show_default=False)],
    exec: Annotated[
        bool,
        typer.Option(
            "--exec/--no-exec",
            help="Exec SSH to connect to the job. (Default: --exec)",
            show_default=False,
        ),
    ] = True,
    show: Annotated[
        ShowOption,
        typer.Option(
            help="What to show: Nothing, SSH Command, Ansible inventory line, "
            "lines for ~/.ssh/config, or the server public key",
            show_default=False,
        ),
    ] = "command",  # pyright: ignore [reportArgumentType]  weird but this works
    proxy: Annotated[ProxyOption, typer.Option(help="Use an SSH jump proxy.", case_sensitive=False)] = "auto",  # pyright: ignore [reportArgumentType]  weird but this works
):
    """Connect to a job using SSH.

    Alternatively, show info, such as the login command.

    Extra arguments are passed directly to ssh.
    """
    # if show in (ShowOption.proxy_host, ShowOption.proxy_pubkey_openssh) and proxy == ProxyOption.auto:
    if show in (ShowOption.proxy_host,) and proxy == ProxyOption.auto:
        # if proxy info is requested, turn on proxy
        proxy = ProxyOption.on

    job, ssh_login, ssh_command_options, destination, ssh_config_jump_proxy = _ssh_common(ctx, job_id, proxy)

    command = ["ssh", *ssh_command_options, destination, *ctx.args]

    command_line = " ".join(command)

    do_exec_ssh = exec and show in [ShowOption.command, ShowOption.nothing]

    if show == ShowOption.command:
        if do_exec_ssh:
            console.print(f"[bold blue]{command_line}")
        else:
            typer.echo(command_line)  # output without any markup
    elif show == ShowOption.ansible:
        # output without any markup
        typer.echo(_job_to_ansible(job, ssh_login, proxy == ProxyOption.on))
    elif show == ShowOption.json:
        obj = {
            "host": ssh_login.host,
            "port": ssh_login.port,
            "username": ssh_login.username,
        }
        if ssh_login.jump_proxy:
            obj["jump_proxy"] = {
                "host": ssh_login.jump_proxy.host,
                "port": ssh_login.jump_proxy.port,
                "username": ssh_login.jump_proxy.username,
            }
        typer.echo(json.dumps(obj, indent=3))

    elif show == ShowOption.ssh_config:
        # output without any markup
        typer.echo(f"""Host={ssh_login.host}
   HostName {ssh_login.host}
   Port {ssh_login.port}
   User {ssh_login.username}
   PreferredAuthentications publickey
   IdentitiesOnly yes
   {ssh_config_jump_proxy}""")
    elif show == ShowOption.server_host:
        typer.echo(ssh_login.host)
    elif show == ShowOption.proxy_host:
        if ssh_login.jump_proxy:
            typer.echo(ssh_login.jump_proxy.host)
        else:
            err_console.print("No jump proxy")
    # elif show == ShowOption.server_pubkey_openssh:
    #     if ssh_login.host_keys:
    #         for host_key in ssh_login.host_keys:
    #             typer.echo(f"{host_key.strip()}")
    #     else:
    #         err_console.print("No hostkeys in SSH info")
    # elif show == ShowOption.proxy_pubkey_openssh:
    #     if ssh_login.jump_proxy:
    #         if ssh_login.jump_proxy.host_keys:
    #             for host_key in ssh_login.jump_proxy.host_keys:
    #                 typer.echo(f"{host_key.strip()}")
    #         else:
    #             err_console.print("No hostkeys in proxy SSH info")
    #     else:
    #         err_console.print("No jump proxy")

    if do_exec_ssh:
        os.execvp(command[0], command)  # noqa: S606


@ai_app.command(
    "sftp",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
    rich_help_panel="Job interaction commands",
)
@handle_some_exceptions
def sftp_to_resource(
    ctx: AIContext,
    job_id: Annotated[str, typer.Argument(help="ID of the job.", show_default=False)],
    exec: Annotated[
        bool,
        typer.Option(
            "--exec/--no-exec",
            help="Exec SFTP to connect to the resource. (Default: --exec)",
            show_default=False,
        ),
    ] = True,
    show_command: Annotated[
        bool,
        typer.Option(
            "--show-command/--no-show-command",
            help="Show command. (Default: --show-command)",
            show_default=False,
        ),
    ] = True,
    proxy: Annotated[ProxyOption, typer.Option(help="Use an SSH jump proxy.", case_sensitive=False)] = "auto",  # pyright: ignore [reportArgumentType]  weird but this works
):
    """Transfer files to/from a job using SFTP.

    The sftp tool is part of openssh and provides an interactive sftp shell.
    Extra arguments are passed directly to sftp.
    """
    job, ssh_login, ssh_command_options, destination, ssh_config_jump_proxy = _ssh_common(
        ctx, job_id, proxy, sftp_mode=True
    )

    command = ["sftp", *ssh_command_options, destination, *ctx.args]

    command_line = " ".join(command)

    if show_command:
        if exec:
            console.print(f"[bold blue]{command_line}")
        else:
            typer.echo(command_line)  # output without any markup

    if exec:
        os.execvp(command[0], command)  # noqa: S606
